/*
 * class for 3dscan projects
 * provides camera transformations
 *
 * transformations copied from libCalibrateCam sources (cal_tran.c by Reg Willson)
 *
 *		Marcel Lancelle 2002
 */

#include "CamTrafo.h"
extern "C" {
//#include "buildOptions.h"
#include "libCamCalib.h"
}
#include <math.h>
#include "mathext.h"

CamTrafo::CamTrafo() {
	f = 1;
	k1 = 0;
	tx = 0;
	ty = 0;
	tz = 0;
	rx = 0;
	ry = 0;
	rz = 0;
	imgWidth = 320;
	imgHeight = 240;
	cx = imgWidth / 2;
	cy = imgHeight / 2;
}

void CamTrafo::copyFrom(CamTrafo* srcPtr) {
	f = srcPtr->f;
	k1 = srcPtr->k1;
	tx = srcPtr->tx;
	ty = srcPtr->ty;
	tz = srcPtr->tz;
	rx = srcPtr->rx;
	ry = srcPtr->ry;
	rz = srcPtr->rz;
	imgWidth = srcPtr->imgWidth;
	imgHeight = srcPtr->imgHeight;
	cx = srcPtr->cx;
	cy = srcPtr->cy;
}

// call this to calculate transformation matrix that is used for most of following methods
void CamTrafo::calcTrafoMatrix() {
	resetMat(&trafo);
	VEC4 t;
	resetVec(&t);
	t.v[0] = tx;
	t.v[1] = ty;
	t.v[2] = tz;

	rotateX(&trafo, rx*180/PI);
	rotateY(&trafo, ry*180/PI);
	rotateZ(&trafo, rz*180/PI);

/*
	float sa = sin(rx);
	float sb = sin(ry);
	float sg = sin(rz);
	float ca = cos(rx);
	float cb = cos(ry);
	float cg = cos(rz);

    trafo.m[0][0] = cb * cg;
    trafo.m[0][1] = cg * sa * sb - ca * sg;
    trafo.m[0][2] = sa * sg + ca * cg * sb;
    trafo.m[1][0] = cb * sg;
    trafo.m[1][1] = sa * sb * sg + ca * cg;
    trafo.m[1][2] = ca * sb * sg - cg * sa;
    trafo.m[2][0] = -sb;
    trafo.m[2][1] = cb * sa;
    trafo.m[2][2] = ca * cb;
*/

	translate(&trafo, &t);

	calcTrafoInverseMatrix();
}

// same but additional rotation around z before other rotations
void CamTrafo::calcTrafoMatrixRotateAxis(VEC4* axis, float angle) {
	VEC4 t;
	resetVec(&t);
	t.v[0] = tx;
	t.v[1] = ty;
	t.v[2] = tz;

//	resetMat(&trafo);
//	rotateZ(&trafo, angle);
	buildRotationMatrix(&trafo, axis, angle);

	rotateX(&trafo, rx*180/PI);
	rotateY(&trafo, ry*180/PI);
	rotateZ(&trafo, rz*180/PI);

	translate(&trafo, &t);
	calcTrafoInverseMatrix();
}

#define SQRT(x)   sqrt(fabs(x))
#define SQR(a)    ((a) * (a))
#define CUB(a)    ((a)*(a)*(a))

// Some math libraries may not have the sincos() routine
#ifdef _SINCOS_
void sincos();
#define SINCOS(x,s,c)   sincos(x,&s,&c)
#else
//double sin(), cos();
#define SINCOS(x,s,c)   s=sin(x);c=cos(x)
#endif

void CamTrafo::calcTrafoInverseMatrix() {
/*
	trafoInverse.m[0][0] = trafo.m[0][0];
	trafoInverse.m[0][1] = trafo.m[1][0];
	trafoInverse.m[0][2] = trafo.m[2][0];
	trafoInverse.m[0][3] = -trafo.m[0][0]*trafo.m[0][3] - trafo.m[1][0]*trafo.m[1][3] - trafo.m[2][0]*trafo.m[2][3];
	trafoInverse.m[1][0] = trafo.m[0][1];
	trafoInverse.m[1][1] = trafo.m[1][1];
	trafoInverse.m[1][2] = trafo.m[2][1];
	trafoInverse.m[1][3] = -trafo.m[0][1]*trafo.m[0][3] - trafo.m[1][1]*trafo.m[1][3] - trafo.m[2][1]*trafo.m[2][3];
	trafoInverse.m[2][0] = trafo.m[0][2];
	trafoInverse.m[2][1] = trafo.m[1][2];
	trafoInverse.m[2][2] = trafo.m[2][2];
	trafoInverse.m[2][3] = -trafo.m[0][2]*trafo.m[0][3] - trafo.m[1][2]*trafo.m[1][3] - trafo.m[2][2]*trafo.m[2][3];
	trafoInverse.m[3][0] = 0;
	trafoInverse.m[3][1] = 0;
	trafoInverse.m[3][2] = 0;
	trafoInverse.m[3][3] = 1;
*/
	getInvRotMat(&trafo, &trafoInverse);
}
#include <windows.h>
#include <stdio.h>
// calculates tx, ty, tz and rx, ry, rz from trafo matrix (upper left 3x3 must be orthonormal rotation matrix)
void CamTrafo::calcParamsFromTrafoMatrix() {
	tx = trafo.m[0][3];
	ty = trafo.m[1][3];
	tz = trafo.m[2][3];

	double sg, cg;
	rz = atan2(trafo.m[1][0], trafo.m[0][0]);
	SINCOS(rz, sg, cg);
	ry = atan2(-trafo.m[2][0], trafo.m[0][0] * cg + trafo.m[1][0] * sg);
	rx = atan2(trafo.m[0][2] * sg - trafo.m[1][2] * cg, trafo.m[1][1] * cg - trafo.m[0][1] * sg);
}

void CamTrafo::world2Img(double xw, double yw, double zw, double* xImg, double* yImg) {
	double xc, yc, zc, xu, yu, Xd, Yd;

	// convert from world coordinates to camera coordinates
	world2Cam(xw, yw, zw, &xc, &yc, &zc);

	// convert from camera coordinates to undistorted image coordinates
	xu = f * xc / zc * 100 + cx; // * 100 are some constant factors
	yu = f * yc / zc * 100 + cy;

	// convert from undistorted to distorted image coordinates
	undistImg2DistImg(xu, yu, xImg, yImg);
}

void CamTrafo::world2Cam(double xw, double yw, double zw, double* xc, double* yc, double* zc) {
	*xc = trafo.m[0][0]*xw + trafo.m[0][1]*yw + trafo.m[0][2]*zw + trafo.m[0][3];
	*yc = trafo.m[1][0]*xw + trafo.m[1][1]*yw + trafo.m[1][2]*zw + trafo.m[1][3];
	*zc = trafo.m[2][0]*xw + trafo.m[2][1]*yw + trafo.m[2][2]*zw + trafo.m[2][3];
}

void CamTrafo::cam2World(double xc, double yc, double zc, double* xw, double* yw, double* zw) {
	*xw = trafoInverse.m[0][0]*xc + trafoInverse.m[0][1]*yc + trafoInverse.m[0][2]*zc + trafoInverse.m[0][3];
	*yw = trafoInverse.m[1][0]*xc + trafoInverse.m[1][1]*yc + trafoInverse.m[1][2]*zc + trafoInverse.m[1][3];
	*zw = trafoInverse.m[2][0]*xc + trafoInverse.m[2][1]*yc + trafoInverse.m[2][2]*zc + trafoInverse.m[2][3];
/*
	double common_denominator;

	// these equations were found by simply inverting the previous routine using Macsyma
	common_denominator = ((trafo.m[0][0] * trafo.m[1][1] - trafo.m[0][1] * trafo.m[1][0]) * trafo.m[2][2] +
			(trafo.m[1][2] * trafo.m[1][0] - trafo.m[0][0] * trafo.m[1][2]) * trafo.m[2][1] +
			(trafo.m[0][1] * trafo.m[1][2] - trafo.m[0][2] * trafo.m[1][1]) * trafo.m[2][0]);

	*xw = ((trafo.m[0][1] * trafo.m[1][2] - trafo.m[0][2] * trafo.m[1][1]) * zc +
		(trafo.m[0][2] * trafo.m[2][1] - trafo.m[0][1] * trafo.m[2][2]) * yc +
		(trafo.m[1][1] * trafo.m[2][2] - trafo.m[1][2] * trafo.m[2][1]) * xc +
		(trafo.m[0][2] * trafo.m[1][1] - trafo.m[0][1] * trafo.m[1][2]) * trafo.m[2][3] +
		(trafo.m[0][1] * trafo.m[2][2] - trafo.m[0][2] * trafo.m[2][1]) * trafo.m[1][3] +
		(trafo.m[1][2] * trafo.m[2][1] - trafo.m[1][1] * trafo.m[2][2]) * trafo.m[0][3]) / common_denominator;

	*yw = -((trafo.m[0][0] * trafo.m[1][2] - trafo.m[0][2] * trafo.m[1][0]) * zc +
		(trafo.m[0][2] * trafo.m[2][0] - trafo.m[0][0] * trafo.m[2][2]) * yc +
		(trafo.m[1][0] * trafo.m[2][2] - trafo.m[1][2] * trafo.m[2][0]) * xc +
		(trafo.m[0][2] * trafo.m[1][0] - trafo.m[0][0] * trafo.m[1][2]) * trafo.m[2][3] +
		(trafo.m[0][0] * trafo.m[2][2] - trafo.m[0][2] * trafo.m[2][0]) * trafo.m[1][3] +
		(trafo.m[1][2] * trafo.m[2][0] - trafo.m[1][0] * trafo.m[2][2]) * trafo.m[0][3]) / common_denominator;

	*zw = ((trafo.m[0][0] * trafo.m[1][1] - trafo.m[0][1] * trafo.m[1][0]) * zc +
		(trafo.m[0][1] * trafo.m[2][0] - trafo.m[0][0] * trafo.m[2][1]) * yc +
		(trafo.m[1][0] * trafo.m[2][1] - trafo.m[1][1] * trafo.m[2][0]) * xc +
		(trafo.m[0][1] * trafo.m[1][0] - trafo.m[0][0] * trafo.m[1][1]) * trafo.m[2][3] +
		(trafo.m[0][0] * trafo.m[2][1] - trafo.m[0][1] * trafo.m[2][0]) * trafo.m[1][3] +
		(trafo.m[1][1] * trafo.m[2][0] - trafo.m[1][0] * trafo.m[2][1]) * trafo.m[0][3]) / common_denominator;
*/
}

void CamTrafo::undistImg2DistImg(double xu, double yu, double* xd, double* yd) {
	double Xu, Yu, Xd, Yd;

	// convert from image to sensor coordinates
	Xu = 0.01 * (xu - cx);
	Yu = 0.01 * (yu - cy);

	// convert from undistorted sensor to distorted sensor plane coordinates
#define SQRT3   1.732050807568877293527446341505872366943
	double Ru, Rd, lambda, c, d, Q, R, D, S, T, sinT, cosT;

	if (((Xu == 0) && (Yu == 0)) || (k1 == 0)) {
		*xd = xu;
		*yd = yu;
		return;
	}

	Ru = hypot (Xu, Yu);	// SQRT(Xu*Xu+Yu*Yu)

	c = 1 / k1;
	d = -c * Ru;

	Q = c / 3;
	R = -d / 2;
	D = CUB (Q) + SQR (R);

	if (D >= 0) {		// one real root
		D = SQRT (D);
		S = CBRT (R + D);
		T = CBRT (R - D);
		Rd = S + T;

		if (Rd < 0) {
			Rd = SQRT (-1 / (3 * k1));
//			fprintf (stderr, "\nWarning: undistorted image point to distorted image point mapping limited by\n");
//			fprintf (stderr, "         maximum barrel distortion radius of %lf\n", Rd);
//			fprintf (stderr, "         (Xu = %lf, Yu = %lf) -> (Xd = %lf, Yd = %lf)\n\n", Xu, Yu, Xu * Rd / Ru, Yu * Rd / Ru);
		}
	} else {			// three real roots
		D = SQRT (-D);
		S = CBRT (hypot (R, D));
		T = atan2 (D, R) / 3;
		SINCOS (T, sinT, cosT);

		// the larger positive root is    2*S*cos(T)
		// the smaller positive root is   -S*cos(T) + SQRT(3)*S*sin(T)
		// the negative root is           -S*cos(T) - SQRT(3)*S*sin(T)

		Rd = -S * cosT + SQRT3 * S * sinT;	// use the smaller positive root
	}

	lambda = Rd / Ru;

	Xd = Xu * lambda;
	Yd = Yu * lambda;

	// convert from sensor to image coordinates
	*xd = Xd * 100 + cx;
	*yd = Yd * 100 + cy;
}

void CamTrafo::distImg2UndistImg(double xd, double yd, double* xu, double* yu) {
	double  Xd, Yd, Xu, Yu;

	// convert from image to sensor coordinates
	Xd = 0.01 * (xd - cx);
	Yd = 0.01 * (yd - cy);

	// convert from distorted sensor to undistorted sensor plane coordinates
	double distortion_factor;
	distortion_factor = 1 + k1 * (SQR (Xd) + SQR (Yd));
	Xu = Xd * distortion_factor;
	Yu = Yd * distortion_factor;

	// convert from sensor to image coordinates */
	*xu = Xu * 100 + cx;
	*yu = Yu * 100 + cy;
}

// returns ray a + m*b (where a is camera position and b points into the scene)
void CamTrafo::img2World(double xImg, double yImg, double* ax, double* ay, double* az, double* bx, double* by, double* bz) {
	// convert from image to distorted sensor coordinates
	double Xd = 0.01 * (xImg - cx);
	double Yd = 0.01 * (yImg - cy);

	// convert from distorted sensor to undistorted sensor plane coordinates
	double distortion_factor;
	distortion_factor = 1 + k1 * (SQR (Xd) + SQR (Yd));
	double Xu = Xd * distortion_factor;
	double Yu = Yd * distortion_factor;

	cam2World(0, 0, 0, ax, ay, az);
	cam2World(Xu, Yu, f, bx, by, bz);

	*bx -= *ax;
	*by -= *ay;
	*bz -= *az;

/*
	// calculate the corresponding xw and yw world coordinates
	// (these equations were derived by simply inverting
	// the perspective projection equations using Macsyma)
	double common_denominator = ((cc.r1 * cc.r8 - cc.r2 * cc.r7) * Yu +
		(cc.r5 * cc.r7 - cc.r4 * cc.r8) * Xu -
		cc.f * cc.r1 * cc.r5 + cc.f * cc.r2 * cc.r4);

	*xw = (((cc.r2 * cc.r9 - cc.r3 * cc.r8) * Yu +
		(cc.r6 * cc.r8 - cc.r5 * cc.r9) * Xu -
		cc.f * cc.r2 * cc.r6 + cc.f * cc.r3 * cc.r5) * zw +
		(cc.r2 * cc.Tz - cc.r8 * cc.Tx) * Yu +
		(cc.r8 * cc.Ty - cc.r5 * cc.Tz) * Xu -
		cc.f * cc.r2 * cc.Ty + cc.f * cc.r5 * cc.Tx) / common_denominator;

	*yw = -(((cc.r1 * cc.r9 - cc.r3 * cc.r7) * Yu +
		(cc.r6 * cc.r7 - cc.r4 * cc.r9) * Xu -
		cc.f * cc.r1 * cc.r6 + cc.f * cc.r3 * cc.r4) * zw +
		(cc.r1 * cc.Tz - cc.r7 * cc.Tx) * Yu +
		(cc.r7 * cc.Ty - cc.r4 * cc.Tz) * Xu -
		cc.f * cc.r1 * cc.Ty + cc.f * cc.r4 * cc.Tx) / common_denominator;
*/
}
