/**
 * class for 3dscan projects
 * that provides camera transformations
 *
 * \author Marcel Lancelle
 * \date 2002-2003
 * \warning transformation 3d world coordinates unit is mm
 */

#ifndef CAMTRAFO_HEADER_INCLUDED
#define CAMTRAFO_HEADER_INCLUDED

#include "math3d.h"

class CamTrafo {
  public:
	CamTrafo();

	/// initialize member variables with values of the argument instance
	void copyFrom(CamTrafo* srcPtr);

	// call one of these to calculate transformation matrix that is used for most of following methods
	/// calculates transformation matrix (in this order: rotation x, y, z; translation) using rx, ry, rz, tx, ty, tz
	void calcTrafoMatrix();
	/// calculates transformation matrix (in this order: rotation axis, x, y, z; translation)
	void calcTrafoMatrixRotateAxis(VEC4* axis, float angle); // same but additional rotation around z before other rotations

	void calcParamsFromTrafoMatrix(); ///< calculates tx, ty, tz and rx, ry, rz from trafo matrix (upper left 3x3 must be orthonormal rotation matrix)

	// transformations; !! world coordinates in millimeter !!
	void world2Img(double xw, double yw, double zw, double* xImg, double* yImg);
	void img2World(double xImg, double yImg, double* ax, double* ay, double* az, double* bx, double* by, double* bz); // returns ray a + m*b (where a is camera position and b points into the scene)

	void world2Cam(double xw, double yw, double zw, double* xc, double* yc, double* zc);
	void cam2World(double xc, double yc, double zc, double* xw, double* yw, double* zw);

	// only distortion correcting transformation
	void undistImg2DistImg(double xu, double yu, double* xd, double* yd);
	void distImg2UndistImg(double xd, double yd, double* xu, double* yu);

	float f; ///< focal length (of thin lens model) [mm]
	float k1; ///< first order lens distortion [1/mm^2]
	float tx; ///< translation x [mm]
	float ty; ///< translation y [mm]
	float tz; ///< translation z [mm]
	float rx; ///< rotation around x axis [rad]
	float ry; ///< rotation around y axis [rad]
	float rz; ///< rotation around z axis [rad]
	float cx; ///< image origin (should be almost center) x [pixel]
	float cy; ///< image origin (should be almost center) y [pixel]
	int imgWidth, imgHeight; ///< [pixel]

  protected:
	void calcTrafoInverseMatrix();

  public:
	MAT4x4 trafo; ///< transformation matrix used for world --> cam coordinate transformations
	MAT4x4 trafoInverse; ///< transformation matrix used for cam --> world coordinate transformations
};

#endif // CAMTRAFO_HEADER_INCLUDED
