/**
 * image class for signal processing and reading and writing bitmap files
 * the image data is stored linearly in memory (in rgb if 24 bit): (0,0), (1,0), ...
 *
 * \bug 8 bit bitmap palette is not respected; gray value images may be misinterpreted if the palette is not a black to white transition (starting with black)
 * \todo bitmap file access and signal processing should be split in 2 classes
 * \todo support for more bitmap formats like 2 colored images or saving palette colors
 * \author Marcel Lancelle
 * \date 2002-2003
 */

#ifndef MBMPIMG_HEADER_INCLUDED
#define MBMPIMG_HEADER_INCLUDED

#include <windows.h>

class MBmpImg {
  public:
	//@{
	/// constructors set width and height to 0, if error occurs
	MBmpImg(int width, int height, char bitsPerPixel);
	MBmpImg(char* bmpFileName);
	MBmpImg(BYTE* bmpPtr); ///< points to BITMAPINFO / BITMAPINFOHEADER
	~MBmpImg();
	//@}

	/// writes to file directly without overhead like creating a class member
	/// bmpPtr points to BITMAPINFO / BITMAPINFOHEADER
	static bool writeBmp2File(BYTE* bmpPtr, char* bmpFileName);
	bool write2BmpFile(char* bmpFileName);
//	BYTE* write2BmpMem();

	int getWidth();
	int getHeight();
	char getBitsPerPixel();

	BYTE* getData(); ///< points to pixel array
	BYTE* getDataXY(int x, int y); ///< points to pixel at position (x,y)
	int getPalSize();
	BYTE* getPalData();

	/// returns linearly interpolated gray value
	//@{
	BYTE getGrayLinIntPolXY(float x, float y);
	void getRGBLinIntPolXY(float x, float y, BYTE* r, BYTE* g, BYTE* b);
	//@}

	/// get the corresponding gray value for a color image
	BYTE getDataXYGrayInRGB(int x, int y);

	/////////////////////////////////////////////////////////////////////
	// image manipulation routines

	static void getHSL(int r, int g, int b, int* h, int* s, int* l); ///< h, s, l: 0..255
	static void getRGB(int h, int s, int l, int* r, int* g, int* b); ///< h, s, l: 0..255

	/// line functions: no clipping -> paint lines only within image borders!!
	//@{
	void lineGray(int x1, int y1, int x2, int y2, BYTE gray);
	void lineRGB(int x1, int y1, int x2, int y2, BYTE r, BYTE g, BYTE b);
	//@}

	void RGB2HSL(MBmpImg* dest);
	void HSL2RGB(MBmpImg* dest);
	void tmMedianRGB(MBmpImg* dest, float radius);
	void medianRGB3x3(MBmpImg* dest); ///< faster than general version
	void sobelRGB(MBmpImg* dest, float factor = 1.0);
	void sobelRGB2Gray(MBmpImg* dest, float factor = 1.0);
	void erodeGray(MBmpImg* dest, int size);
	void dilateGray(MBmpImg* dest, int size);
	void convolute(MBmpImg* tmpImg, float* factors, int num); // convolutes rows and columns with given 1d signal
	void blurGaussian121(MBmpImg* tmpImg);

	/// label (enumerate) regions of 8 bit image (non-zero pixels; regions must not touch image borders!)
	/// there are 2 versions using a different bit resolution for the temporary enumeration
	/// \return number of labels found
	//@{
	int label8(); // work on same 8 bit image
	int label16(); // use temporary 16 bit image
	//@}

	void fillHisto(int *histoPtr); ///< fills an array like int histo[256] with fillHisto(&histo[0])

  protected:
	BYTE* dataPtr;
	char bitsPerPixel;
	int width, height;

	int rowSize;
	BYTE* palPtr; ///< points to palette table
	int palSize; ///< number of palette entries
};

typedef MBmpImg* MBmpImgPtr;

#endif // MBMPIMG_HEADER_INCLUDED
