#include "MTwain.h"

// string declarations for log file (as in Twain Specification 1.9)
const char* const MTwain::TEXT_STATE[8] = {
	"undefined",
	"pre session (1)",
	"source manager loaded (2)",
	"source manager open (3)",
	"source open (4)",
	"source enabled (5)",
	"transfer ready (6)",
	"transferring (7)"
};

const char* const MTwain::TEXT_RC[TEXT_RC_SIZE] = {
	"TWRC_SUCCESS (0)",
	"TWRC_FAILURE (1)",
	"TWRC_CHECKSTATUS (2)",
	"TWRC_CANCEL (3)",
	"TWRC_DSEVENT (4)",
	"TWRC_NOTDSEVENT (5)",
	"TWRC_XFERDONE (6)",
	"TWRC_ENDOFLIST (7)",
	"TWRC_INFONOTSUPPORTED (8)",
	"TWRC_DATANOTAVAILABLE (9)"
};

const char* const MTwain::TEXT_CC[TEXT_CC_SIZE] = {
	"TWCC_SUCCESS (0)",
	"TWCC_BUMMER (1)",
	"TWCC_LOWMEMORY (2)",
	"TWCC_NODS (3)",
	"TWCC_MAXCONNECTIONS (4)",
	"TWCC_OPERATIONERROR (5)",
	"TWCC_BADCAP (6)",
	"? (7)",
	"? (8)",
	"TWCC_BADPROTOCOL (9)",
	"TWCC_BADVALUE (10)",
	"TWCC_SEQERROR (11)",
	"TWCC_BADDEST (12)",
	"TWCC_CAPUNSUPPORTED (13)",
	"TWCC_CAPBADOPERATION (14)",
	"TWCC_CAPSEQERROR (15)",
	"TWCC_DENIED (16)",
	"TWCC_FILEEXISTS (17)",
	"TWCC_FILENOTFOUND (18)",
	"TWCC_NOTEMPTY (19)",
	"TWCC_PAPERJAM (20)",
	"TWCC_PAPERDOUBLEFEED (21)",
	"TWCC_FILEWRITEERROR (22)",
	"TWCC_CHECKDEVICEONLINE (23)"
};

MTwain::MTwain() {
	state = STATE_PRE_SESSION;
	logFile = NULL;
}

MTwain::~MTwain() {
	log("destructor called\n");
	if (state == STATE_PRE_SESSION) return;

	if (state == STATE_SOURCE_MANAGER_OPEN) closeSourceManager();

	if (state != STATE_SOURCE_MANAGER_LOADED) {
		// error case: try to clean up
		log("WARNING: destructor called in state %s\n", (char*) TEXT_STATE[state]);

		backToState(STATE_SOURCE_MANAGER_LOADED);
	}

	// releases Twain DLL
	FreeLibrary(TwainDLL);
	setState(STATE_PRE_SESSION);

	stopLog();
}

// methods controlling logging to a file
//
// every time startLog is called the log file is overwritten
// typically call startLog before init, stopLog is called by destructor automatically
// or if you debug a small piece of code, put it between startLog and stopLog
bool MTwain::startLog(char* logFileName) {
	logFile = fopen(logFileName, "w");
	//if ((int) logFile == -1) logFile = NULL; // ????
	if (logFile == NULL) return false;
	log("log file of TWAIN functions (see TWAIN specification or source code for more info)\n\n");
	return true;
}
void MTwain::stopLog() {
	if (logFile) fclose(logFile);
}
// adds a string to log file
inline void MTwain::log(char* text) {
	if (logFile) fprintf(logFile, text);
}
// adds a string with int parameter to log file
inline void MTwain::log(char* text, int val) {
	if (logFile) fprintf(logFile, text, val);
}
// adds a string with string parameter to log file
inline void MTwain::log(char* text, char* val) {
	if (logFile) fprintf(logFile, text, val);
}

// loads Twain DLL and gets DSM Entry
// state: STATE_PRE_SESSION --> STATE_SOURCE_MANAGER_LOADED
bool MTwain::init(HWND hwnd) {
	log("init called\n");
	if (state != STATE_PRE_SESSION) {
		log(" WARNING: MTwain is already initialized\n");
		return false;
	}

	if (!IsWindow(hwnd)) {
		log(" ERROR: hwnd must exist when MTwain::init(hwnd) is called\n");
		return false;
	}
	this->hwnd = hwnd;

	// try to load the TWAIN DLL
	// make it so Windows doesn't complain if it isn't found
	unsigned int oldErrorMode = SetErrorMode(SEM_NOOPENFILEERRORBOX);
	TwainDLL = LoadLibrary("TWAIN_32.DLL");
	SetErrorMode(oldErrorMode);	

	if (TwainDLL == NULL) return false;

	// Load the main entry point DSM_ENTRY
	DSMEntryProc = (DSMENTRYPROC) GetProcAddress(TwainDLL, MAKEINTRESOURCE(1));
	if (DSMEntryProc == NULL) {
		FreeLibrary(TwainDLL);
		return false;
	}

	setState(STATE_SOURCE_MANAGER_LOADED);
	return true;
}

///////////////////////////////////////////////////////////////////////////////
///// low level methods
///////////////////////////////////////////////////////////////////////////////

// helper function for callDS and callDSM; source manager must allready be loaded
void MTwain::callTwainEntry(pTW_IDENTITY appIdPtr, pTW_IDENTITY srcIdPtr,
		TW_UINT32 dataGroup, TW_UINT16 dataArgument,
		TW_UINT16 messageID, TW_MEMREF dataPtr) {

	// call entry proc
	lastReturnCode = (*DSMEntryProc) (appIdPtr, srcIdPtr, dataGroup, dataArgument,
			messageID, dataPtr);

	// don't log preProcessMessage calls
	if (messageID == MSG_PROCESSEVENT) {
		lastStatusConditionCode = TWCC_SUCCESS;
		return;
	}

	if (lastReturnCode < TEXT_RC_SIZE) log(" return code: %s\n", (char*) TEXT_RC[lastReturnCode]);
		else log(" WARNING: return code %i unknown", lastReturnCode);

	if (lastReturnCode == TWRC_SUCCESS) {
		lastStatusConditionCode = TWCC_SUCCESS;
		return;
	}

	// getting more info by checking status condition code
	TW_STATUS twStatus;
	if ((*DSMEntryProc) (&appIdentity, NULL, DG_CONTROL, DAT_STATUS,
			MSG_GET, &twStatus)) {
		lastStatusConditionCode = twStatus.ConditionCode;
		if (lastStatusConditionCode < TEXT_CC_SIZE) log(" status condition code: %s\n", (char*) TEXT_CC[lastStatusConditionCode]);
			else log(" WARNING: status condition code %i unknown\n", lastStatusConditionCode);
	} else {
		lastStatusConditionCode = TWCC_BUMMER;
		log(" ERROR: status condition code couldn't be read successfully\n");
	}
}

// dataGroup: DG_CONTROL, DG_IMAGE, DG_AUDIO
// dataArgument: DAT_xxxx
// messageID: MSG_xxxx (MSG_GET, MSG_SET, ...)
// dataPtr: pointer to data structure (in most cases you have to allocate the mem, in all cases you have to deallocate it!)
bool MTwain::callDS(TW_UINT32 dataGroup, TW_UINT16 dataArgument,
					TW_UINT16 messageID, TW_MEMREF dataPtr) {
	if (messageID != MSG_PROCESSEVENT) log("callDS\n"); // don't log preProcessMessage calls
	if (state < STATE_SOURCE_OPEN) {
		log(" ERROR: current state is %s but source must be open to work with it\n", (char*) TEXT_STATE[state]);
		return false;
	}

	callTwainEntry(&appIdentity, &srcIdentity, dataGroup, dataArgument,
			messageID, dataPtr);

	return (lastReturnCode == TWRC_SUCCESS);
}

bool MTwain::callDSM(TW_UINT32 dataGroup, TW_UINT16 dataArgument,
					TW_UINT16 messageID, TW_MEMREF dataPtr) {
	log("callDSM\n");
	if (state < STATE_SOURCE_MANAGER_LOADED) {
		log(" ERROR: current state is %s but source manager must be loaded in order to call it\n", (char*) TEXT_STATE[state]);
		return false;
	}

	callTwainEntry(&appIdentity, NULL, dataGroup, dataArgument,
			messageID, dataPtr);

	return (lastReturnCode == TWRC_SUCCESS);
}

// if callDS or callDSM are not sucessful call this to get more info
TW_UINT16 MTwain::getLastReturnCode() {
	return lastReturnCode;
}

// if callDS or callDSM is not sucessful call this to get more info; returns TWCC_xxxx
TW_UINT16 MTwain::getLastStatusConditionCode() {
	return lastStatusConditionCode;
}

char MTwain::getState() {
	return state;
}

void MTwain::setState(char newState) {
	char oldState = state;
	state = newState;
	onStateChange(oldState, newState);
}

// try to go back to state; use this in error case
// state -> newState
void MTwain::backToState(char newState) {
	log("backToState called; trying to recover...\n");
	if (state > STATE_SOURCE_ENABLED) {
		setState(STATE_SOURCE_ENABLED);
	}
	if (newState >= STATE_SOURCE_ENABLED) return;

	// try to disable data source
	if (state > STATE_SOURCE_OPEN) {
		TW_USERINTERFACE ui; // not used (but why there is a pointer to it?)
		callDS(DG_CONTROL, DAT_USERINTERFACE, MSG_DISABLEDS, (TW_MEMREF) &ui);
		setState(STATE_SOURCE_OPEN);
	}
	if (newState == STATE_SOURCE_OPEN) return;

	// try to close data source
	callDSM(DG_CONTROL, DAT_IDENTITY, MSG_CLOSEDS, (TW_MEMREF) &srcIdentity);
	setState(STATE_SOURCE_MANAGER_OPEN);
	if (newState == STATE_SOURCE_MANAGER_OPEN) return;

	// try to close source manager
	callDSM(DG_CONTROL, DAT_PARENT, MSG_CLOSEDSM, (TW_MEMREF) &hwnd);
	setState(STATE_SOURCE_MANAGER_LOADED);
	if (newState == STATE_SOURCE_MANAGER_LOADED) return;

	// try to releases Twain DLL
	FreeLibrary(TwainDLL);
	setState(newState);
}

///////////////////////////////////////////////////////////////////////////////
///// medium level methods
///////////////////////////////////////////////////////////////////////////////

// returns true if message was processed
// must be called from message loop, normally of the main application
bool MTwain::preProcessMessage(LPMSG msgPtr) {
	if (state < STATE_SOURCE_ENABLED) return false; // only when data source is enabled

	// determine if event belongs to data source
	TW_EVENT twEvent;
	twEvent.pEvent = (TW_MEMREF) msgPtr;
	twEvent.TWMessage = MSG_NULL;
	callDS(DG_CONTROL, DAT_EVENT, MSG_PROCESSEVENT, (TW_MEMREF) &twEvent);

	// if not: return false so that main message loop of application can process message
	if (lastReturnCode != TWRC_DSEVENT) return false;

	// if yes: process it here
	switch(twEvent.TWMessage) {
		case MSG_XFERREADY: // data is ready for transfer
			log("message MSG_XFERREADY received\n");
			setState(STATE_TRANSFER_READY);
			transferData();
			break;
		case MSG_CLOSEDSREQ: // request that the source's user interface be disabled
			log("message MSG_CLOSEDSREQ received\n");
			onCloseDSRequest();
			break;
		case MSG_CLOSEDSOK: // request that the source's user interface be disabled
							// (special case for DG_CONTROL, DAT_USERINTERFACE, MSG_ENABLEDSUIONLY)
			log("message MSG_CLOSEDSOK received\n");
			break;
		case MSG_DEVICEEVENT: // device event has occured
			// there may be many device events, so they aren't written to log file
			break;
	}

	return true;
}

// fills appIdentity and opens source manager
// state: STATE_SOURCE_MANAGER_LOADED --> STATE_SOURCE_MANAGER_OPEN
bool MTwain::openSourceManager(char* appName, char* manufacturer,
		TW_UINT16 language /*= TWLG_ENG*/, TW_UINT16 country /*= TWCY_GERMANY*/,
		TW_UINT16 versionMajor /*= 1*/, TW_UINT16 versionMinor /*= 0*/) {

	log("openSourceManager\n");
	if (state != STATE_SOURCE_MANAGER_LOADED) {
		log(" ERROR: called in state %s\n", (char*) TEXT_STATE[state]);
		return false;
	}

	// fill appIdentity
	appIdentity.Id = 0; // is assigned by source manager during call
	appIdentity.Version.MajorNum = versionMajor;
	appIdentity.Version.MinorNum = versionMinor;
	appIdentity.Version.Language = language;
	appIdentity.Version.Country = country;
	lstrcpy(appIdentity.Version.Info, "");
	appIdentity.ProtocolMajor = TWON_PROTOCOLMAJOR;
	appIdentity.ProtocolMinor = TWON_PROTOCOLMINOR;
	appIdentity.SupportedGroups = DG_IMAGE | DG_CONTROL;
	lstrcpy(appIdentity.Manufacturer, manufacturer);
	lstrcpy(appIdentity.ProductFamily, "");
	lstrcpy(appIdentity.ProductName, appName);

	// open source manager
	if (!callDSM(DG_CONTROL, DAT_PARENT, MSG_OPENDSM, (TW_MEMREF) &hwnd)) return false;

	setState(STATE_SOURCE_MANAGER_OPEN);

	selectDefaultSource(); // reset srcIdentity

	return true;
}

// resets srcIdentity so that succeding open source call will open default source
// you normally don't call this unless you want to deselect a custom selected souce e.g.
// 'default source' is either the last one used by the user or the last one installed
// state: STATE_SOURCE_MANAGER_OPEN
void MTwain::selectDefaultSource() {
	log("selectDefaultSource\n");
	if (state != STATE_SOURCE_MANAGER_OPEN) {
		log(" ERROR: called in state %s\n", (char*) TEXT_STATE[state]);
		return;
	}

	// fill srcIdentity
	srcIdentity.Id = 0;
	lstrcpy(srcIdentity.ProductName, "");
}

// open dialog with all sources
// state: STATE_SOURCE_MANAGER_OPEN
bool MTwain::selectSourceDialog() {
	log("selectSourceDialog\n ");
	if (state != STATE_SOURCE_MANAGER_OPEN) {
		log(" ERROR: called in state %s\n", (char*) TEXT_STATE[state]);
		return false;
	}

	selectDefaultSource(); // reset srcIdentity

	// show dialog
	return callDSM(DG_CONTROL, DAT_IDENTITY, MSG_USERSELECT, (TW_MEMREF) &srcIdentity);
}

// opens data source (default source if none special source was selected before)
// state: STATE_SOURCE_MANAGER_OPEN --> STATE_SOURCE_OPEN
bool MTwain::openSource() {
	log("openSource\n");
	if (state != STATE_SOURCE_MANAGER_OPEN) {
		log(" ERROR: called in state %s\n", (char*) TEXT_STATE[state]);
		return false;
	}

	// open data source
	if (!callDSM(DG_CONTROL, DAT_IDENTITY, MSG_OPENDS, (TW_MEMREF) &srcIdentity)) return false;
	setState(STATE_SOURCE_OPEN);
	return true;
}

// gets Capability, fill in desired capability before (cap.Cap = CAP_xxxx)
// delete cap.hContainer (which is allocated by source) yourself when you don't
// need it any more! (see Twain Spec!)
// state: >= STATE_SOURCE_OPEN (typical: = STATE_SOURCE_OPEN)
bool MTwain::getCapability(pTW_CAPABILITY capPtr) {
	log("getCapability\n");
	if (state < STATE_SOURCE_OPEN) {
		log(" ERROR: called in state %s\n", (char*) TEXT_STATE[state]);
		return false;
	}

	// get capability
	return callDS(DG_CONTROL, DAT_CAPABILITY, MSG_GET, (TW_MEMREF) capPtr);
}

// sets Capability; returns false if cap couldn't be set to the exact value
//	cap.Cap = TWCP_xxxx
//	cap.ConType = TWON_xxxx (TWON_ARRAY, TWON_ENUMERATION, TWON_ONEVALUE, TWON_RANGE)
/*
vorgeschlagen in Spec:
	twCapability.hContainer = GlobalAlloc(GHND, sizeof(TW_ONEVALUE));
	pval = (pTW_ONEVALUE) GlobalLock(twCapability.hContainer);
	pval->ItemType = TWTY_INT16;
	pval->Item = 1; //This app will only accept 1 image
	GlobalUnlock(twCapability.hContainer);
	...
	GlobalFree((HANDLE)twContainer.hContainer);

mein Vorschlag (wei nicht, ob der geht!):
	TW_ONEVALUE capParam;
	twCapability.hContainer = &capParam;
	capParam.ItemType = TWTY_INT16;
	capParam.Item = 1;
	...
*/
// state: >= STATE_SOURCE_OPEN (typical: = STATE_SOURCE_OPEN)
bool MTwain::setCapability(pTW_CAPABILITY capPtr) {
	log("setCapability\n");
	if (state < STATE_SOURCE_OPEN) {
		log(" ERROR: called in state %s\n", (char*) TEXT_STATE[state]);
		return false;
	}

	// set capability
	return callDS(DG_CONTROL, DAT_CAPABILITY, MSG_SET, (TW_MEMREF) capPtr);
	// TWRC_SUCCESS: exact value could be set
	// TWRC_CHECKSTATUS: the value could not be matched exactly
	// TWRC_FAILURE: see condition code:
		// TWCC_BADCAP, TWCC_CAPUNSUPPORTED, TWCC_CAPBADOPERATION, TWCC_CAPSEQERROR: Source does not support setting this cap
		// TWCC_BADDEST: The Source specified by pSourceID is not open
		// TWCC_BADVALUE: The value set was out of range for this Source: Use MSG_GET to determine what setting was made (see the TWRC_CHECKSTATUS case)
		// TWCC_SEQERROR: Operation invoked in invalid state
}

// enables data source (requests acquisition)
// state: STATE_SOURCE_OPEN --> STATE_SOURCE_ENABLED
bool MTwain::enableSource(bool showUI, bool modalUI) {
	log("enableSource\n");
	if (state != STATE_SOURCE_OPEN) {
		log(" ERROR: called in state %s\n", (char*) TEXT_STATE[state]);
		return false;
	}

	// setup ui struct
	TW_USERINTERFACE ui;
	ui.ShowUI = showUI;
	ui.ModalUI = modalUI;
	ui.hParent = hwnd;

	// enable data source
	if (!callDS(DG_CONTROL, DAT_USERINTERFACE, MSG_ENABLEDS, (TW_MEMREF) &ui)) return false;
	setState(STATE_SOURCE_ENABLED);
	return true;
}

// transfers image data (overwrite this if you use audio); is called by preProcessMessage()
// state: STATE_TRANSFER_READY --> STATE_TRANSFERRING
bool MTwain::transferData() {
	log("transferData\n");
	if (state != STATE_TRANSFER_READY) {
		log(" ERROR: called in state %s\n", (char*) TEXT_STATE[state]);
		return false;
	}

	// get image info
	if (!callDS(DG_IMAGE, DAT_IMAGEINFO, MSG_GET, (TW_MEMREF) &imageInfo)) return false;

	// start transfer
	setState(STATE_TRANSFERRING);
	TW_UINT32 bmpHandle;
	callDS(DG_IMAGE, DAT_IMAGENATIVEXFER, MSG_GET, (TW_MEMREF) &bmpHandle);

	switch (lastReturnCode) {
		case TWRC_XFERDONE:
			setState(STATE_TRANSFER_READY);
			imageDataPtr = (BYTE *) GlobalLock((HANDLE) bmpHandle);
			onImageTransferDone(imageInfo.ImageWidth, imageInfo.ImageLength, imageInfo.BitsPerPixel);
			GlobalFree((HANDLE) bmpHandle);

			TW_PENDINGXFERS pendTransfers;
			if (!callDS(DG_CONTROL, DAT_PENDINGXFERS, MSG_ENDXFER, (TW_MEMREF) &pendTransfers)) return false;
			if (pendTransfers.Count == 0) setState(STATE_SOURCE_ENABLED); // otherwise stays in state STATE_TRANSFER_READY
			return true;

		case TWRC_CANCEL:
		case TWRC_FAILURE:
			// clean up incomplete transfer
			log(" incomplete transfer\n");
			setState(STATE_SOURCE_ENABLED);
			return false;
	}
	return false;
}

// disables data source; causes hiding of user interface if it was shown
// state: STATE_SOURCE_ENABLED --> STATE_SOURCE_OPEN
bool MTwain::disableSource() {
	log("disableSource\n");
	if (state != STATE_SOURCE_ENABLED) {
		log(" ERROR: called in state %s\n", (char*) TEXT_STATE[state]);
		return false;
	}

	// disable data source
	TW_USERINTERFACE ui; // not used (but why there is a pointer to it?)
	if (!callDS(DG_CONTROL, DAT_USERINTERFACE, MSG_DISABLEDS, (TW_MEMREF) &ui)) return false;
	setState(STATE_SOURCE_OPEN);
	return true;
}

// closes data source
// state: STATE_SOURCE_OPEN --> STATE_SOURCE_MANAGER_OPEN
bool MTwain::closeSource() {
	log("disableSource\n");
	if (state != STATE_SOURCE_OPEN) {
		log(" ERROR: called in state %s\n", (char*) TEXT_STATE[state]);
		return false;
	}

	// close data source
	if (!callDSM(DG_CONTROL, DAT_IDENTITY, MSG_CLOSEDS, (TW_MEMREF) &srcIdentity)) return false;
	setState(STATE_SOURCE_MANAGER_OPEN);
	return true;
}

// closes source manager
// state: STATE_SOURCE_MANAGER_OPEN --> STATE_SOURCE_MANAGER_LOADED
bool MTwain::closeSourceManager() {
	log("closeSourceManager\n");
	if (state != STATE_SOURCE_MANAGER_OPEN) {
		log(" ERROR: called in state %s\n", (char*) TEXT_STATE[state]);
		return false;
	}

	// close source manager
	if (!callDSM(DG_CONTROL, DAT_PARENT, MSG_CLOSEDSM, (TW_MEMREF) &hwnd)) return false;

	setState(STATE_SOURCE_MANAGER_LOADED);
	return true;
}

// shows user interface for configuration purpose
// state: STATE_SOURCE_OPEN --> STATE_SOURCE_ENABLED
bool MTwain::showConfigSourceUI() {
	log("show ui for configuration\n");
	if (state != STATE_SOURCE_OPEN) {
		log(" ERROR: called in state %s\n", (char*) TEXT_STATE[state]);
		return false;
	}

	TW_USERINTERFACE ui; // not used (but why there is a pointer to it?)
//log(" -> bef\n");
	/* TODO (#1#): doesn't work - why? */
	if (!callDS(DG_CONTROL, DAT_USERINTERFACE, MSG_ENABLEDSUIONLY, (TW_MEMREF) &ui)) return false;

	MSG msg;
	while (GetMessage((LPMSG)&msg, NULL, 0, 0)) {
		if (!preProcessMessage(&msg)) {
			TranslateMessage ((LPMSG)&msg);
			DispatchMessage ((LPMSG)&msg);
		}
	}

//log(" -> aft\n");
	setState(STATE_SOURCE_ENABLED);
	return disableSource();
}

///////////////////////////////////////////////////////////////////////////////
///// declaration virtual methods (perhaps you don't always implement them)
///////////////////////////////////////////////////////////////////////////////

// is called when internal class state changes
// use it for example for changing cursor / icon during transfer or for debugging
//
// beware that from state 4 to 5 if you show a MessageBox e.g. perhaps messages
// as MSG_XFERREADY will get lost during the MessageBox is shown, that's really
// important if you don't show the UserInterface and wait for an image to be transferred
void MTwain::onStateChange(char oldState, char newState) {
	if (logFile) fprintf(logFile, "state transition: %s --> %s\n", TEXT_STATE[oldState], TEXT_STATE[newState]);
}

// is called when image transfer complete: implement this to get image data
// you find the image data at class variable 'imageDataPtr'
void MTwain::onImageTransferDone(int width, int height, char bitsPerPixel) {
}

// is called when data source wants to be disabled; standard implementation disables and closes data source
void MTwain::onCloseDSRequest() {
	disableSource();
	closeSource();
}

///////////////////////////////////////////////////////////////////////////////
///// high level methods
///////////////////////////////////////////////////////////////////////////////

// state before: STATE_PRESESSION or STATE_SOURCE_MANAGER_LOADED
// state after (if successfull): STATE_SOURCE_MANAGER_OPEN
bool MTwain::initialize(HWND hwnd, char* appName, char* manufacturer) {
	if (state == STATE_PRE_SESSION) {
		if (!init(hwnd)) return false;
	}

	if (state > STATE_SOURCE_MANAGER_LOADED) backToState(STATE_SOURCE_MANAGER_LOADED);

	if (!openSourceManager(appName, manufacturer)) return false;
}

// state: STATE_SOURCE_MANAGER_OPEN
void MTwain::showSelectSourceDialog() {
	if (getState() == STATE_SOURCE_MANAGER_OPEN) selectSourceDialog();
		else MessageBox(hwnd, "can't show 'select source dialog'", "unexpected sequence", MB_ICONEXCLAMATION | MB_OK);
}

void MTwain::startStandardAcquisition() {
	openSource();
	enableSource(true, false);
}

bool MTwain::startAcquisitionNoUI() {
	openSource();
	return enableSource(false, false);
/*
	if (!openSource()) return false;
	if (!enableSource(false, false)) {
		closeSource();
		return false;
	}
	return true;
*/
}

void MTwain::stopAcquisition() {
	disableSource();
	closeSource();
}

// is source capable of CAP_xxxx?
bool MTwain::getCapabilityBool(TW_UINT16 cap) {
	TW_CAPABILITY twCap;
	twCap.Cap = cap;
	twCap.ConType = TWON_DONTCARE16; // source will specify
	twCap.hContainer = NULL; // source allocates and fills container

	if (!getCapability(&twCap)) return false;

	pTW_ONEVALUE valPtr;
	valPtr = (pTW_ONEVALUE) GlobalLock(twCap.hContainer);
	bool ret = valPtr->Item;
	GlobalFree((HANDLE)twCap.hContainer); //free hContainer allocated by Source

	return ret;
}
