/**
 * class providing basic TWAIN access
 *
 * uses twain.h (that I modified to work with mingw)
 *
 * still does support only native image transfer
 *
 *
 * \author Marcel Lancelle
 * \date 2002-2003
 */

#ifndef MTWAIN_HEADER_INCLUDED

#include <windows.h>
#include "twain.h" // include windows.h before twain.h!
#include <stdio.h>

class MTwain {
  public:
	MTwain();
	~MTwain();

	///////////////////////////////////////////////////////////////////////////////
	///// high level methods

	/// state before: STATE_PRESESSION or STATE_SOURCE_MANAGER_LOADED
	/// state after (if successfull): STATE_SOURCE_MANAGER_OPEN
	bool initialize(HWND hwnd, char* appName, char* manufacturer);

	/// state: STATE_SOURCE_MANAGER_OPEN
	void showSelectSourceDialog();

	void startStandardAcquisition();
	bool startAcquisitionNoUI();
	void stopAcquisition();

/*
 * most of methods below for more special twain access when deriving your class
 *
 * refer to TWAIN specification at www.twain.org and you will easily be
 * able to use these methods
*/

	///////////////////////////////////////////////////////////////////////////////
	///// medium level methods

	/// returns true if message was processed
	/// must be called from message loop, normally of the main application
	bool preProcessMessage(LPMSG msgPtr);

	/// methods controlling logging to a file
	///
	/// every time startLog is called the log file is overwritten
	/// typically call startLog before init, stopLog is called by destructor automatically
	/// or if you debug a small piece of code, put it between startLog and stopLog
	//@{
	bool startLog(char* logFileName);
	void stopLog();
	//@}

	/// is source capable of CAP_xxxx?
	bool getCapabilityBool(TW_UINT16 cap);

	///////////////////////////////////////////////////////////////////////////////
	/// following states as described in TWAIN specification
	//@{
	static const char STATE_PRE_SESSION = 1;
	static const char STATE_SOURCE_MANAGER_LOADED = 2;
	static const char STATE_SOURCE_MANAGER_OPEN = 3;
	static const char STATE_SOURCE_OPEN = 4;
	static const char STATE_SOURCE_ENABLED = 5;
	static const char STATE_TRANSFER_READY = 6;
	static const char STATE_TRANSFERRING = 7;
	//@}

  protected:
	///////////////////////////////////////////////////////////////////////////////
	///// virtual methods

	/// is called when image transfer complete: implement this to get image data
	/// you find the image data at class variable 'imageDataPtr'
	virtual void onImageTransferDone(int width, int height, char bitsPerPixel);

	/// is called when internal class state changes
	/// use it for example for changing cursor / icon during transfer or for debugging
	///
	/// beware that from state 4 to 5 if you show a MessageBox e.g. perhaps messages
	/// as MSG_XFERREADY will get lost during the MessageBox is shown, that's really
	/// important if you don't show the UserInterface and wait for an image to be transferred
	virtual void onStateChange(char oldState, char newState);

	/// is called when data source wants to be disabled; standard implementation disables and closes data source
	virtual void onCloseDSRequest();

	///////////////////////////////////////////////////////////////////////////////
	///// low level methods

	/// dataGroup: DG_CONTROL, DG_IMAGE, DG_AUDIO
	/// dataArgument: DAT_xxxx
	/// messageID: MSG_xxxx (MSG_GET, MSG_SET, ...)
	/// dataPtr: pointer to data structure (in most cases you have to allocate the mem, in all cases you have to deallocate it!)
	//@{
	bool callDS(TW_UINT32 dataGroup, TW_UINT16 dataArgument,
					TW_UINT16 messageID, TW_MEMREF dataPtr);
	bool callDSM(TW_UINT32 dataGroup, TW_UINT16 dataArgument,
					TW_UINT16 messageID, TW_MEMREF dataPtr);
	//@}

	/// if callDS or callDSM are not sucessful call this to get more info
	TW_UINT16 getLastReturnCode();

	/// if callDS or callDSM is not sucessful call this to get more info; returns TWCC_xxxx
	TW_UINT16 getLastStatusConditionCode();
	char getState();
	void setState(char newState);

	/// try to go back to state; use this in error case
	void backToState(char newState);

	/// adds a string to log file
	void log(char* text);
	/// adds a string with int parameter to log file
	void log(char* text, int val);
	/// adds a string with string parameter to log file
	void log(char* text, char* val);

	///////////////////////////////////////////////////////////////////////////////
	///// medium level methods

	/// loads Twain DLL and gets DSM Entry
	/// state: STATE_PRE_SESSION --> STATE_SOURCE_MANAGER_LOADED
	bool init(HWND hwnd);

	/// fills appIdentity and opens source manager
	/// state: STATE_SOURCE_MANAGER_LOADED --> STATE_SOURCE_MANAGER_OPEN
	bool openSourceManager(char* appName, char* manufacturer,
		TW_UINT16 language = TWLG_ENG, TW_UINT16 country = TWCY_GERMANY,
		TW_UINT16 versionMajor = 1, TW_UINT16 versionMinor = 0);

	/// resets srcIdentity so that succeding open source call will open default source
	/// you normally don't call this unless you want to deselect a custom selected souce e.g.
	/// 'default source' is either the last one used by the user or the last one installed
	/// state: STATE_SOURCE_MANAGER_OPEN
	void selectDefaultSource();

	/// open dialog with all sources
	/// state: STATE_SOURCE_MANAGER_OPEN
	bool selectSourceDialog();

	/// opens data source (default source if none special source was selected before)
	/// state: STATE_SOURCE_MANAGER_OPEN --> STATE_SOURCE_OPEN
	bool openSource();

	/// gets Capability, fill in desired capability before (cap.Cap = CAP_xxxx)
	/// delete cap.hContainer (which is allocated by source) yourself when you don't
	/// need it any more! (see Twain Spec!)
	/// state: >= STATE_SOURCE_OPEN (typical: = STATE_SOURCE_OPEN)
	bool getCapability(pTW_CAPABILITY capPtr);

	/// sets Capability; returns false if cap couldn't be set to the exact value
	///	cap.Cap = TWCP_xxxx
	///	cap.ConType = TWON_xxxx (TWON_ARRAY, TWON_ENUMERATION, TWON_ONEVALUE, TWON_RANGE)
	/// state: >= STATE_SOURCE_OPEN (typical: = STATE_SOURCE_OPEN)
// UNTESTED !!
	bool setCapability(pTW_CAPABILITY capPtr);

	/// enables data source (requests acquisition)
	/// state: STATE_SOURCE_OPEN --> STATE_SOURCE_ENABLED
	bool enableSource(bool showUI, bool modalUI);

	/// transfers image data (overwrite this if you use audio); is called by preProcessMessage()
	/// state: STATE_TRANSFER_READY --> STATE_TRANSFERRING
	virtual bool transferData();

	/// disables data source; causes hiding of user interface if it was shown
	/// state: STATE_SOURCE_ENABLED --> STATE_SOURCE_OPEN
	bool disableSource();

	/// closes data source
	/// state: STATE_SOURCE_OPEN --> STATE_SOURCE_MANAGER_OPEN
	bool closeSource();

	/// closes source manager
	/// state: STATE_SOURCE_MANAGER_OPEN --> STATE_SOURCE_MANAGER_LOADED
	bool closeSourceManager();

	/// shows user interface for configuration purpose
	/// state: STATE_SOURCE_OPEN --> STATE_SOURCE_ENABLED
	bool showConfigSourceUI();

	///////////////////////////////////////////////////////////////////////
	// member variables

	TW_IDENTITY appIdentity;
	TW_IDENTITY srcIdentity;
	TW_IMAGEINFO imageInfo;
	BYTE* imageDataPtr;
	HWND hwnd;
	FILE* logFile;

	/// string definitions for log file
	//@{
	static const char* const TEXT_STATE[8];
	static const int TEXT_RC_SIZE = 10;
	static const char* const TEXT_RC[TEXT_RC_SIZE];
	static const int TEXT_CC_SIZE = 24;
	static const char* const TEXT_CC[TEXT_CC_SIZE];
	//@}

  private:
	/// helper function for callDS and callDSM
	void callTwainEntry(pTW_IDENTITY appIdPtr, pTW_IDENTITY srcIdPtr,
		TW_UINT32 dataGroup, TW_UINT16 dataArgument,
		TW_UINT16 messageID, TW_MEMREF dataPtr);

	char state; // see above

	TW_UINT16 lastReturnCode;
	TW_UINT16 lastStatusConditionCode;

	HINSTANCE TwainDLL;
	DSMENTRYPROC DSMEntryProc;
};

#define MTWAIN_HEADER_INCLUDED
#endif // MTWAIN_HEADER_INCLUDED
