/*
 * Marching Cubes
 *
 * modified implementation (not producing holes)
 *
 * At ambigous faces always negative vertices are connected using
 * a variable triangulation. This leads to a consistent topology
 * and a manifold object.
 * Additional incorrect edges on faces are avoided without using centroids.
 * The maximum number of triangles per cube is only 5.
 *
 * by
 *  Cedric Vigne
 *  Marcel Lancelle
 *
 * input:
 *  samples of type char (surface at 0: outside < 0 < inside)
 *  holes are produced when object is clipped so make sure you have
 *  surrounding negative values around your object
 *
 * output:
 *  obj file
 *
 *
 *    6______ 7
 *    /     /|      z
 *  4/_____/ |      |
 *   |    5| |      |   y
 *   |     | /3     | /
 *   |_____|/       |/_____ x
 *  0      1
 *
 *  vertex coding:
 *  2 | 1 | 0
 *  |   |   |--> 0: -x    1: +x
 *  |   |------> 0: -y    1: +y
 *  |----------> 0: -z    1: +z
 *
 *  configuration coding:
 *  7 | ... | 0
 *  |         |--> vertex 000 positive
 *  |------------> vertex 111 positive
 *
 */

#include "MarchingCubes.h"
#include <io.h>
#include <windows.h>
#include <gl/gl.h>

static const int confOfMainConf[15] = {0, 1, 3, 33, 129, 14, 131, 146, 15, 77, 153, 141, 30, 105, 78};
// structure definissant nos 256 configurations
struct confStruct {
	int mainConfNum; // base config of which it's a symmetric case
	short rotation[3][3]; // coords = rotation matrix * base config coords
	bool invertedPM; // more as 4 of the 8 vertices above threshold? (-> reversed order of triangle vertices)
} conf[256];

MarchingCubes::MarchingCubes() {
	initConfTable();
}

void MarchingCubes::init(int sizeX, int sizeY, int sizeZ,
		float scaleX /* = 1.0 */, float scaleY /* = 1.0 */, float scaleZ /* = 1.0 */) {
	this->sizeX = sizeX;
	this->sizeY = sizeY;
	this->sizeZ = sizeZ;

	// calc translation and scaling for final coords:
	// coords = (arrayIndex - transl) * tScale
	translX = (sizeX-1) * 0.5;
	translY = (sizeY-1) * 0.5;
	translZ = (sizeZ-1) * 0.5;
	this->tScaleX = scaleX * 2.0 / (sizeX-1);
	this->tScaleY = scaleY * 2.0 / (sizeY-1);
	this->tScaleZ = scaleZ * 2.0 / (sizeZ-1);

	// reserve and reset merging tables
	for (int i=0; i<3; i++)	{
		vertexMergingTable[i] = new int[2*sizeX * 2*sizeY];
		if (!vertexMergingTable[i]) throw "not enough memory";

		for (int y=0; y<2*sizeY; y++)
			for (int x=0; x<2*sizeX; x++) {
				vertexMergingTable[i][x + y*2*sizeX] = -1;
			}
	}
}


// set internal factors
void MarchingCubes::translate(float translX/*=0.0*/, float translY/*=0.0*/, float translZ/*=0.0*/) {
	this->translX += translX;
	this->translY += translY;
	this->translZ += translZ;
}

// initializes configurations: calculates index, rotation matrix
// and inverted flag for each possible config
void MarchingCubes::initConfTable() {
	for (int i=0; i<256; i++) conf[i].mainConfNum = -1;
	
	int configFoundNum = 0; // for speeding up initialization

	for (int turnX=0; turnX<4; turnX++) {
		for (int turnY=0; turnY<4; turnY++) {
			for (int turnZ=0; turnZ<4; turnZ++) {
				// for each rotation ...
				for (int i=0; i<15; i++) {
					// ... generate rotated configurations for each main configuration i
					unsigned int turnedConf = confOfMainConf[i];
					short rot[3][3]; // temporary rotation matrix
					// reset rot to identity
					for (int j=0; j<3; j++) for (int k=0; k<3; k++) rot[j][k] = (j == k)?1:0;
		
					// rotate around x-axis, 90 degrees for each rotation
					for (int j=1; j<=turnX; j++) {
						// exchange bits to obtain rotated configuration
						turnedConf = ( (((turnedConf & 0x20)>0)<<7)
								| (((turnedConf & 0x10)>0)<<6)
								| (((turnedConf & 0x02)>0)<<5)
								| (((turnedConf & 0x01)>0)<<4)
								| (((turnedConf & 0x80)>0)<<3)
								| (((turnedConf & 0x40)>0)<<2)
								| (((turnedConf & 0x08)>0)<<1)
								| ((turnedConf & 0x04)>0) );

						// rotate temporary rotation matrix
						short t;
						for (int p=0; p<3; p++) {
							t = rot[p][2];
							rot[p][2] = -rot[p][1];
							rot[p][1] = t;
						}
					}
					
					// rotate around y-axis, 90 degrees for each rotation
					for (int j=1; j<=turnY; j++) {
						// exchange bits to obtain rotated configuration
						turnedConf = ( (((turnedConf & 0x08)>0)<<7)
								| (((turnedConf & 0x80)>0)<<6)
								| (((turnedConf & 0x02)>0)<<5)
								| (((turnedConf & 0x20)>0)<<4)
								| (((turnedConf & 0x04)>0)<<3)
								| (((turnedConf & 0x40)>0)<<2)
								| (((turnedConf & 0x01)>0)<<1)
								| ((turnedConf & 0x10)>0) );

						// rotate temporary rotation matrix
						short t;
						for (int p=0; p<3; p++) {
							t = rot[p][0];
							rot[p][0] = -rot[p][2];
							rot[p][2] = t;
						}
					}
					
					// rotate around z-axis, 90 degrees for each rotation
					for (int j=1; j<=turnZ; j++) {
						// exchange bits to obtain rotated configuration
						turnedConf = ( (((turnedConf & 0x40)>0)<<7)
								| (((turnedConf & 0x10)>0)<<6)
								| (((turnedConf & 0x80)>0)<<5)
								| (((turnedConf & 0x20)>0)<<4)
								| (((turnedConf & 0x04)>0)<<3)
								| (((turnedConf & 0x01)>0)<<2)
								| (((turnedConf & 0x08)>0)<<1)
								| ((turnedConf & 0x02)>0) );

						// rotate temporary rotation matrix
						short t;
						for (int p=0; p<3; p++) {
							t = rot[p][1];
							rot[p][1] = -rot[p][0];
							rot[p][0] = t;
						}
					}
					
					// write to lookup table
					if (conf[turnedConf].mainConfNum == -1) {
						conf[turnedConf].mainConfNum = i;
						configFoundNum++;
						for (int j=0; j<3; j++) for (int k=0; k<3; k++) conf[turnedConf].rotation[j][k] = rot[j][k];
						conf[turnedConf].invertedPM = false;
					}

					unsigned int turnedConfInv = (~turnedConf)&0xFF;
					//fprintf(stderr,"turnConfinv=%i\n",turnedConfInv);
					// write symmetric configruration to lookup table
					if (conf[turnedConfInv].mainConfNum == -1) {
						conf[turnedConfInv].mainConfNum = i;
						configFoundNum++;
						for (int j=0; j<3; j++) for (int k=0; k<3; k++) conf[turnedConfInv].rotation[j][k] = rot[j][k];
						conf[turnedConfInv].invertedPM = true;
					}

					if (configFoundNum > 255) return; // all configs already found; exit method
				}
			}
		}
	}
}

void MarchingCubes::processSlice(int z0, char sliceZ0[], char sliceZ1[]) {
	// set data pointers
	this->sliceZ[0] = sliceZ0;
	this->sliceZ[1] = sliceZ1;

	// prepare vertex merging tables
	// exchange tables 0 and 2 so that last array is first now
	int* tmpXPtr = vertexMergingTable[0];
	vertexMergingTable[0] = vertexMergingTable[2];
	vertexMergingTable[2] = tmpXPtr;
	// reset values in tables 1 and 2
	for (int i=1; i<3; i++)	{
		for (int y=0; y<2*sizeY; y++)
			for (int x=0; x<2*sizeX; x++) {
				vertexMergingTable[i][x + y*2*sizeX] = -1;
			}
	}

	for (int y=0; y<sizeY-1; y++)
		for (int x=0; x<sizeX-1; x++) {
                	generateTriangles(x, y, z0);
		}
}

// to be called after the loop with processSlice
void MarchingCubes::terminate() {
	// deletes temporary vertex merging tables
	for (int i=0; i<3; i++)	delete[] vertexMergingTable[i];
}

// draws model (OpenGL)
// !!! slow !!! doesn't use neither vertex arrays nor display lists yet
void MarchingCubes::draw() {
	glBegin(GL_TRIANGLES);
	for (int t=0; t<triangles.size(); t++) { // all triangles
		for (int v=0; v<3; v++) { // all three vertices
			int ind = triangles[t].t[v];
			glVertex3f(coords[ind].c[0], coords[ind].c[1], coords[ind].c[2]);
		}
	}
	glEnd();
}

// !!! very slow !!!
void MarchingCubes::drawFlatShaded() {
	for (int t=0; t<triangles.size(); t++) { // all triangles
	glBegin(GL_TRIANGLES);

		// calc and set normal
		float a[3], b[3], n[3];
		int ind0 = triangles[t].t[0];
		int ind1 = triangles[t].t[1];
		int ind2 = triangles[t].t[2];
		for (int i=0; i<3; i++) {
			a[i] = coords[ind1].c[i] - coords[ind0].c[i];
			b[i] = coords[ind2].c[i] - coords[ind0].c[i];
		}
		n[0] = a[1]*b[2] - a[2]*b[1];
		n[1] = a[2]*b[0] - a[0]*b[2];
		n[2] = a[0]*b[1] - a[1]*b[0];
		float l = sqrt(n[0]*n[0] + n[1]*n[1] + n[2]*n[2]);
		for (int i=0; i<3; i++) n[i] /= l;
		glNormal3fv(&n[0]);

		for (int v=0; v<3; v++) { // all three vertices
			int ind = triangles[t].t[v];
			glVertex3f(coords[ind].c[0], coords[ind].c[1], coords[ind].c[2]);
		}

	glEnd();
/*
glDisable(GL_LIGHTING);
glDisable(GL_CULL_FACE);
glEnable(GL_POLYGON_OFFSET_LINE);
glPolygonOffset(-1.0, -1.0);
glBegin(GL_LINE_LOOP);
for (int v=0; v<3; v++) { // all three vertices
int ind = triangles[t].t[v];
glVertex3f(coords[ind].c[0], coords[ind].c[1], coords[ind].c[2]);
}
glEnd();
glEnable(GL_LIGHTING);
glEnable(GL_CULL_FACE);
glDisable(GL_POLYGON_OFFSET_LINE);
*/
	}
}

// writes Model to the .obj ASCII file format
bool MarchingCubes::write2ObjFile(char* objFileName) {
	FILE* objFile;
	objFile = fopen(objFileName, "w");
	if (objFile == NULL) return false;

	fprintf(objFile, "# 3DScanner Marching Cubes output\n\n");

	// vertices
	for (int i=0; i<coords.size(); i++) {
		fprintf(objFile, "v %f %f %f\n", coords[i].c[0], coords[i].c[1], coords[i].c[2]);
	}

	fprintf(objFile, "\n");

	// triangle faces
	for (int i=0; i<triangles.size(); i++) { // all triangles
		fprintf(objFile, "f %i %i %i\n", triangles[i].t[0]+1, triangles[i].t[1]+1, triangles[i].t[2]+1);
	}

	fclose(objFile);
	return true;
}

// writes Model to the .wrl VRML file format (VRML97)
bool MarchingCubes::write2VRMLFile(char* vrmlFileName) {
	FILE* vrmlFile;
	vrmlFile = fopen(vrmlFileName, "w");
	if (vrmlFile == NULL) return false;

	fprintf(vrmlFile, "#VRML V2.0 utf8\n\n");
	fprintf(vrmlFile, "#3DScanner Marching Cubes output\n\n");

	fprintf(vrmlFile, "Shape {\n");
	fprintf(vrmlFile, " geometry IndexedFaceSet {\n");
	fprintf(vrmlFile, "  solid TRUE\n");
	fprintf(vrmlFile, "  ccw TRUE\n");
	fprintf(vrmlFile, "  coord Coordinate {\n");
	fprintf(vrmlFile, "   point [\n");

	// vertices
	for (int i=0; i<coords.size(); i++) {
		fprintf(vrmlFile, "%f %f %f,\n", coords[i].c[0], coords[i].c[1], coords[i].c[2]);
	}

	fprintf(vrmlFile, "   ]  # end point\n");
	fprintf(vrmlFile, "  } # end coord\n");

	// triangle faces: index list
	fprintf(vrmlFile, "  coordIndex [\n");
	for (int i=0; i<triangles.size(); i++) { // all triangles
		fprintf(vrmlFile, "%i, %i, %i, -1,\n", triangles[i].t[0], triangles[i].t[1], triangles[i].t[2]);
	}
	fprintf(vrmlFile, "  ] # end coordIndex\n");


	fprintf(vrmlFile, " } # end geometry\n");
	fprintf(vrmlFile, "} # end shape\n");

	fclose(vrmlFile);
	return true;
}

// generates triangles for one cube
void MarchingCubes::generateTriangles(int x, int y, int z) {
	int numTris = 0;	// nombre de triangle dans la configuration
//	MarchingCubesTriangle tri[4];	//tableau de triangles: dans une configuration, il y a au plus 4 triangles
	MarchingCubesTriangle tri[5];	//tableau de triangles: dans une configuration, il y a au plus 4 triangles

	// calculate configuration for actual cube
	int config = (((sliceZ[1][x+1 + sizeX * (y+1)] > 0)?1:0)<<7)
		+ (((sliceZ[1][x + sizeX * (y+1)] > 0)?1:0)<<6)
		+ (((sliceZ[1][x+1 + sizeX * y] > 0)?1:0)<<5)
		+ (((sliceZ[1][x + sizeX * y] > 0)?1:0)<<4)
		+ (((sliceZ[0][x+1 + sizeX * (y+1)] > 0)?1:0)<<3)
		+ (((sliceZ[0][x + sizeX * (y+1)] > 0)?1:0)<<2)
		+ (((sliceZ[0][x+1 + sizeX * y] > 0)?1:0)<<1)
		+ (((sliceZ[0][x + sizeX * y] > 0)?1:0));

	// return if nothing to do
	if (conf[config].mainConfNum == 0) return;

	// fill list of triangles (vertices are ordered counterclockwise for OpenGL backface culling)
	switch (conf[config].mainConfNum) {
		case  1:
			numTris = 1;
			tri[0].setTriangleVertices(0, -1, -1,   -1, 0, -1,   -1, -1, 0);
			break;
		case  2:
			numTris = 2;
			tri[0].setTriangleVertices(-1, 0, -1,   -1, -1, 0,   +1, -1, 0);
			tri[1].setTriangleVertices(+1, 0, -1,   -1, 0, -1,   +1, -1, 0);
			break;
		case  3:
			if (!conf[config].invertedPM) { // original config
				numTris = 2;
				tri[0].setTriangleVertices(0, -1, -1,   -1, 0, -1,   -1, -1, 0);
				tri[1].setTriangleVertices(0, -1, +1,   +1, 0, +1,   +1, -1, 0);
			} else {
				numTris = 4;
				tri[0].setTriangleVertices(0, -1, +1,   +1, 0, +1,   -1, 0, -1);
				tri[1].setTriangleVertices(0, -1, +1,   -1, 0, -1,   -1, -1, 0);
				tri[2].setTriangleVertices(+1, -1, 0,   0, -1, -1,   -1, 0, -1);
				tri[3].setTriangleVertices(+1, -1, 0,   -1, 0, -1,   +1, 0, +1);
			}
			break;
		case  4:
			numTris = 2;
			tri[0].setTriangleVertices(0, -1, -1,   -1, 0, -1,   -1, -1, 0);
			tri[1].setTriangleVertices(+1, 0, +1,   0, +1, +1,   +1, +1, 0);
			break;
		case  5:
			numTris = 3;
			tri[0].setTriangleVertices(+1, -1, 0,   -1, 0, -1,   0, -1, -1);
			tri[1].setTriangleVertices(+1, -1, 0,   -1, +1, 0,   -1, 0, -1);
			tri[2].setTriangleVertices(+1, -1, 0,   +1, +1, 0,   -1, +1, 0);
			break;
		case  6:
			if (!conf[config].invertedPM) { // original config
				numTris = 3;
				tri[0].setTriangleVertices(+1, 0, +1,   0, +1, +1,   +1, +1, 0);
				tri[1].setTriangleVertices(-1, 0, -1,   -1, -1, 0,   +1, -1, 0);
				tri[2].setTriangleVertices(+1, 0, -1,   -1, 0, -1,   +1, -1, 0);
			} else {
				numTris = 5;
				tri[0].setTriangleVertices(0, +1, +1,   +1, +1, 0,   +1, 0, -1);
				tri[1].setTriangleVertices(0, +1, +1,   +1, 0, -1,   -1, 0, -1);
				tri[2].setTriangleVertices(0, +1, +1,   -1, 0, -1,   -1, -1, 0);
				tri[3].setTriangleVertices(0, +1, +1,   -1, -1, 0,   +1, -1, 0);
				tri[4].setTriangleVertices(0, +1, +1,   +1, -1, 0,   +1, 0, +1);
			}
			break;
		case  7:
			if (!conf[config].invertedPM) { // original config
				numTris = 3;
				tri[0].setTriangleVertices(+1, 0, +1,   0, +1, +1,   +1, +1, 0);
				tri[1].setTriangleVertices(-1, 0, +1,   0, -1, +1,   -1, -1, 0);
				tri[2].setTriangleVertices(+1, -1, 0,   +1, 0, -1,   0, -1, -1);
			} else {
				numTris = 5;
				tri[0].setTriangleVertices(0, -1, -1,   -1, 0, +1,   +1, +1, 0);
				tri[1].setTriangleVertices(0, -1, -1,   +1, +1, 0,   +1, 0, -1);
				tri[2].setTriangleVertices(+1, +1, 0,   -1, 0, +1,   0, +1, +1);
				tri[3].setTriangleVertices(-1, 0, +1,   0, -1, -1,   -1, -1, 0);
				tri[4].setTriangleVertices(+1, -1, 0,   +1, 0, +1,   0, -1, +1);
			}
			break;
		case  8:
			numTris = 2;
			tri[0].setTriangleVertices(+1, +1, 0,   -1, +1, 0,   -1, -1, 0);
			tri[1].setTriangleVertices(+1, +1, 0,   -1, -1, 0,   +1, -1, 0);
			break;
		case  9:
			numTris = 4;
			tri[0].setTriangleVertices(0, +1, +1,   -1, 0, +1,   -1, -1, 0);
			tri[1].setTriangleVertices(0, +1, +1,   -1, -1, 0,   0, -1, -1);
			tri[2].setTriangleVertices(+1, +1, 0,   0, +1, +1,   0, -1, -1);
			tri[3].setTriangleVertices(+1, +1, 0,   0, -1, -1,   +1, 0, -1);
			break;
		case 10:
			if (!conf[config].invertedPM) { // original config
				numTris = 4;
				tri[0].setTriangleVertices(-1, 0, +1,   0, -1, +1,   -1, 0, -1);
				tri[1].setTriangleVertices(0, -1, +1,   0, -1, -1,   -1, 0, -1);
				tri[2].setTriangleVertices(+1, 0, +1,   0, +1, +1,   0, +1, -1);
				tri[3].setTriangleVertices(+1, 0, +1,   0, +1, -1,   +1, 0, -1);
			} else {
				numTris = 4;
				tri[0].setTriangleVertices(0, -1, +1,   0, -1, -1,   +1, 0, -1);
				tri[1].setTriangleVertices(0, -1, +1,   +1, 0, -1,   +1, 0, +1);
				tri[2].setTriangleVertices(0, +1, -1,   -1, 0, +1,   0, +1, +1);
				tri[3].setTriangleVertices(0, +1, -1,   -1, 0, -1,   -1, 0, +1);
			}
			break;
		case 11:
			numTris = 4;
			tri[0].setTriangleVertices(+1, 0, -1,   +1, 0, +1,   0, -1, -1);
			tri[1].setTriangleVertices(+1, 0, +1,   -1, +1, 0,   0, -1, -1);
			tri[2].setTriangleVertices(+1, 0, +1,   0, +1, +1,   -1, +1, 0);
			tri[3].setTriangleVertices(-1, +1, 0,   -1, -1, 0,   0, -1, -1);
			break;
		case 12:
			if (!conf[config].invertedPM) { // original config
				numTris = 4;
				tri[0].setTriangleVertices(-1, 0, +1,   0, -1, +1,   -1, -1, 0);
				tri[1].setTriangleVertices(+1, -1, 0,   -1, 0, -1,   0, -1, -1);
				tri[2].setTriangleVertices(+1, -1, 0,   -1, +1, 0,   -1, 0, -1);
				tri[3].setTriangleVertices(+1, -1, 0,   +1, +1, 0,   -1, +1, 0);
			} else {
				numTris = 4;
				tri[0].setTriangleVertices(0, -1, -1,   -1, -1, 0,   -1, 0, -1);
				tri[1].setTriangleVertices(-1, +1, 0,   -1, 0, +1,   0, -1, +1);
				tri[2].setTriangleVertices(-1, +1, 0,   0, -1, +1,   +1, -1, 0);
				tri[3].setTriangleVertices(-1, +1, 0,   +1, -1, 0,   +1, +1, 0);
			}
			break;
		case 13:
			if (!conf[config].invertedPM) { // original config
				numTris = 4;
				tri[0].setTriangleVertices(0, -1, -1,   -1, 0, -1,   -1, -1, 0);
				tri[1].setTriangleVertices(0, -1, +1,   +1, 0, +1,   +1, -1, 0);
				tri[2].setTriangleVertices(0, +1, +1,   -1, 0, +1,   -1, +1, 0);
				tri[3].setTriangleVertices(+1, 0, -1,   +1, +1, 0,   0, +1, -1);
			} else {
				numTris = 4;
				tri[0].setTriangleVertices(-1, -1, 0,   0, -1, +1,   -1, 0, +1);
				tri[1].setTriangleVertices(0, -1, -1,   +1, 0, -1,   +1, -1, 0);
				tri[2].setTriangleVertices(+1, +1, 0,   0, +1, +1,   +1, 0, +1);
				tri[3].setTriangleVertices(-1, 0, -1,   -1, +1, 0,   0, +1, -1);
			}
			break;
		case 14:
			numTris = 4;
			tri[0].setTriangleVertices(+1, +1, 0,   0, -1, -1,   +1, -1, 0);
			tri[1].setTriangleVertices(+1, +1, 0,   -1, 0, +1,   0, -1, -1);
			tri[2].setTriangleVertices(+1, +1, 0,   0, +1, +1,   -1, 0, +1);
			tri[3].setTriangleVertices(-1, 0, +1,   -1, 0, -1,   0, -1, -1);
			break;
	}

	// apply rotation matrix
	for (int i=0; i<numTris; i++) // for each triangle
		for (int j=0; j<3; j++) { // for each vertex
			char tx = tri[i].loc[j].x;
			char ty = tri[i].loc[j].y;
			char tz = tri[i].loc[j].z;

			tri[i].loc[j].x = conf[config].rotation[0][0]*tx + conf[config].rotation[1][0]*ty + conf[config].rotation[2][0]*tz;
			tri[i].loc[j].y = conf[config].rotation[0][1]*tx + conf[config].rotation[1][1]*ty + conf[config].rotation[2][1]*tz;
			tri[i].loc[j].z = conf[config].rotation[0][2]*tx + conf[config].rotation[1][2]*ty + conf[config].rotation[2][2]*tz;
		}

	// calculate interpolated points for each vertex of triangle
	for (int i=0; i<numTris; i++) { // for each triangle
		MCTrianglesStruct t;
		for (int j=0; j<3; j++) { // for each vertex
			MCCoordsStruct c;

			// test si le point existe: dans le cas contraire, on le cree
			// indices for vertex merging table
			int vmX = 2*x+tri[i].loc[j].x+1;
			int vmY = 2*y+tri[i].loc[j].y+1;
			int vmZ = tri[i].loc[j].z+1;
			int indXY = vmX + vmY*2*sizeX;

			int ptInd; // vertex index for triangle list

			if ((ptInd = vertexMergingTable[vmZ][indXY]) < 0) {
				// point not yet in vertexMergingTable

				// calculate model (final) coordinates
				// si (p=x,y ou z) loc.p==0, cela signifie que sur l'axe p, le point a creer se trouve entre 2 points du cube initial
				int tmp1,tmp2;
				if (tri[i].loc[j].x == 0) {
					tmp1 = ((tri[i].loc[j].y == 1) ? 1 : 0) + y;
					tmp2 = (tri[i].loc[j].z == 1) ? 1 : 0;
					tri[i].interpolatePoint(j,getValue(x,tmp1,tmp2),getValue(x+1,tmp1,tmp2),x,tmp1,tmp2+z,x+1,tmp1,tmp2+z);
				}

				if (tri[i].loc[j].y == 0) {
					tmp1 = ((tri[i].loc[j].x == 1) ? 1 : 0) + x;
					tmp2 = (tri[i].loc[j].z == 1) ? 1 : 0;
					tri[i].interpolatePoint(j,getValue(tmp1,y,tmp2),getValue(tmp1,y+1,tmp2),tmp1,y,tmp2+z,tmp1,y+1,tmp2+z);
				}

				if (tri[i].loc[j].z == 0) {
					tmp1 = ((tri[i].loc[j].x == 1) ? 1 : 0) + x;
					tmp2 = ((tri[i].loc[j].y == 1) ? 1 : 0) + y;
					tri[i].interpolatePoint(j,getValue(tmp1,tmp2,0),getValue(tmp1,tmp2,1),tmp1,tmp2,z,tmp1,tmp2,z+1);
				}

				// add new point to point list and entry to vertexMergingTable
				ptInd = coords.size();
			 	vertexMergingTable[vmZ][indXY] = ptInd;
				c.c[0] = (tri[i].fin[j].x - translX) * tScaleX;
				c.c[1] = (tri[i].fin[j].y - translY) * tScaleY;
				c.c[2] = (tri[i].fin[j].z - translZ) * tScaleZ;
				//printf("c:%.1f %.1f %.1f %i   ", c.c[0], c.c[1], c.c[2], (tri[i].loc[j].z+1)/2+z);
				coords.push_back(c);
			} // else already contained in vertexMergingTable -> use existing point

			int k = ( (conf[config].invertedPM) ? 2-j : j ); // reverse order?
			t.t[k] = ptInd; // assign triangle vertex index
		}
		triangles.push_back(t);
	}
}

// easy access to sliceZ[]: range: 0..sizeX-1, 0..sizeY-1, 0..1
float MarchingCubes::getValue(int x, int y, int z) {
	return sliceZ[z][y*sizeX + x];
}

///////////////////////////////////////////////////////////////
///////////////// class MarchingCubesTriangle /////////////////
///////////////////////////////////////////////////////////////

MarchingCubesTriangle::MarchingCubesTriangle() {
}

void MarchingCubesTriangle::setLoc(int i, int x, int y, int z){
	loc[i].x = x;
	loc[i].y = y;
	loc[i].z = z;
}

void MarchingCubesTriangle::setTriangleVertices(char v0x, char v0y, char v0z,
		char v1x, char v1y, char v1z,
		char v2x, char v2y, char v2z) {
	setLoc(0,v0x,v0y,v0z);
	setLoc(1,v1x,v1y,v1z);
	setLoc(2,v2x,v2y,v2z);	
}
/*
void MarchingCubesTriangle::setFin(int &i,float &x,float &y, float &z){
	fin[i].x = x;
	fin[i].y = y;
	fin[i].z = z;
}
*/
void MarchingCubesTriangle::interpolatePoint(int i, float v1, float v2, float p1x, float p1y, float p1z, float p2x, float p2y, float p2z){
	float absDiff = fabs(v1 - v2);
	float weight2 = (absDiff == 0) ? 0.5 : (fabs(v1) / absDiff); 
	float weight1 = (float)(1.0 - weight2);

	fin[i].x = weight1 * p1x + weight2 * p2x;
	fin[i].y = weight1 * p1y + weight2 * p2y;
	fin[i].z = weight1 * p1z + weight2 * p2z;
}
