/**
 * Marching Cubes
 *
 * modified implementation (not producing holes)
 *
 * At ambigous faces always negative vertices are connected using
 * a variable triangulation. This leads to a consistent topology
 * and a manifold object.
 * Additional incorrect edges on faces are avoided without using centroids.
 * The maximum number of triangles per cube is only 5.
 *
 * \author Cedric Vigne
 * \author Marcel Lancelle
 * \date 2002
 *
 * input:
 *  samples of type char (surface at 0: outside < 0 < inside)
 *  holes are produced when object is clipped so make sure you have
 *  surrounding negative values around your object
 *
 * output:
 *  arrays with vertex coordinates and triangle indices
 *
 */

#ifndef MARCHINGCUBES_HEADER_INCLUDED
#define MARCHINGCUBES_HEADER_INCLUDED

#include <math.h>
#include <stdlib.h>
#include <vector>
using namespace std;

class MarchingCubes {
  public:
	MarchingCubes();

	///\param size: number of samples at each side of the volume
	///\param scale: scaling factor; 1 means range -1 ... 1
	void init(int sizeX, int sizeY, int sizeZ,
		float scaleX=1.0, float scaleY=1.0, float scaleZ=1.0);

	/// translates final object coordinates
	/// object coords = (arrayIndex - transl) * tScale
	void translate(float translX=0.0, float translY=0.0, float translZ=0.0);

	/// you can process in slices with increasing z for saving memory
	///\param slices: char slice[sizeY][sizeX]
	void processSlice(int z0, char sliceZ0[], char sliceZ1[]);
	void terminate(); ///< to be called after the loop with processSlice
/*
	// you can process the whole volume in one step, but all data has to be
	// in memory at a time
	// data: char data[sizeZ][sizeY][sizeX]
	void processVolume(char* data[][][]);
*/

	/// draws model (OpenGL)
	void draw();
	void drawFlatShaded();

	/// writes Model to the .obj ASCII file format
	bool write2ObjFile(char* objFileName);
	/// writes Model to the .wrl VRML file format (VRML97)
	bool write2VRMLFile(char* vrmlFileName);

	/// object coords = (arrayIndex - transl) * tScale
	float tScaleX, tScaleY, tScaleZ;
	/// object coords = (arrayIndex - transl) * tScale
	float translX, translY, translZ;

	struct MCTrianglesStruct {
		int t[3];
	};

	struct MCCoordsStruct {
		float c[3];
	};

	/// contains vertice indices of calculated triangles
	vector<MCTrianglesStruct> triangles;
	/// contains coordinates of calculated triangles
	vector<MCCoordsStruct> coords;
  protected:
	/// initializes rotation matrices for configs
	void initConfTable();

	/// generates triangles for one cube
	void generateTriangles(int x, int y, int z);

	/// easy access to sliceZ[]: range: 0..sizeX-1, 0..sizeY-1, 0..1
	float getValue(int x, int y, int z);

	int sizeX, sizeY, sizeZ;

	/// data pointers to arrays
	char* sliceZ[2];

	/// for using same point definition in all triangles
	int* vertexMergingTable[3];
}; // class MarchingCubes

//////////////////////////////////////////////////////////////////
/// class for triangles used in MarchingCubes::generateTriangles()
class MarchingCubesTriangle {
  public:
	MarchingCubesTriangle();

	struct shortVertexStruct {
		char x, y, z;
	} loc[3]; ///< local coordinates; origin is center of cube

	struct finalVertexStruct{
		float x, y, z;
	} fin[3]; ///< final object coordinates; origin is center of object

	void setLoc(int i, int x, int y, int z); ///< sets cube local coordinates
//	void setFin(int &i, float &x, float &y, float &z);

	// creation du triangle
	void setTriangleVertices(char v0x, char v0y, char v0z,
		char v1x, char v1y, char v1z,
		char v2x, char v2y, char v2z);

	void interpolatePoint(int i, float v1, float v2, float p1x, float p1y, float p1z, float p2x, float p2y, float p2z);
}; // class MarchingCubesTriangle

#endif // MARCHINGCUBES_HEADER_INCLUDED
