#include "ScannerApp.h"
#include "PortAccess.h"
#include <math.h>
#include "mathext.h"
#include "frontend.h"
#include "MLInputDialog.h"
#include "segmentation.h"
#include "multiResImg.h"
#include <string>

#define INI_FILE_NAME		"3DScanner.ini"
//////////////////////////////////////////////////
#define INI_PORTADDRESS		"PortAddress"
#define INI_STEPNUM360		"StepNum360"
#define INI_ROTATEMANUALLY	"RotateManually"
#define INI_DEBUGOUTPUT		"DebugOutput"
#define INI_SEGLOWTHRESHOLD "SegmentationLowThreshold"
#define INI_SEGHIGHTHRESHOLD "SegmentationHighThreshold"

#define INI_IMGAQUISTYPE	"ImageAquisitionType"
#define INI_IAT_FILE		"file"
#define INI_IAT_TWAIN		"twain"
#define INI_IAT_VFW			"vfw"

#define INI_CALIBPATTERN	"CalibrationPattern"
#define INI_CP_1			"1cm"
#define INI_CP_2			"2cm"
//////////////////////////////////////////////////

void ScannerApp::initGl() {
	glClearColor(0.0f, 0.0f, 0.0f, 0.5f);						// Black Background

	// blending
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	// lighting
	GLfloat light_ambient[] = {0.2, 0.2, 0.2, 1.0};
	GLfloat light_diffuse[] = {0.8, 0.8, 0.8, 1.0};
	GLfloat light_specular[] = {1.0, 1.0, 1.0, 1.0};
	// light_position is NOT default value
	GLfloat light0_position[] = {2.0, 1.0, 4.0, 0.0};

	glLightfv(GL_LIGHT0, GL_AMBIENT, light_ambient);
	glLightfv(GL_LIGHT0, GL_DIFFUSE, light_diffuse);
	glLightfv(GL_LIGHT0, GL_SPECULAR, light_specular);
	glLightfv(GL_LIGHT0, GL_POSITION, light0_position);

//	GLfloat color_specular[] = {1.0, 1.0, 1.0, 1.0};
//	GLfloat color_ambient[] = {0.9, 0.2, 0.2, 1.0};
//	GLfloat color_diffuse[] = {0.8, 0.8, 0.8, 1.0};

//	glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, color_specular);
//	glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, color_ambient);
//	glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, color_diffuse);
	glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, 40.0);

	glEnable(GL_LIGHT0);
	glEnable(GL_DEPTH_TEST);
}

void ScannerApp::init(HWND hwnd, HINSTANCE hInstance) {
	this->hwnd = hwnd;
	this->hInstance = hInstance;
	project = NULL;
	mergeProject = NULL;

	showGrid = true;
	showTexture = true;
	showSmoothShaded = true;
	showMesh = false;
	showNormals = false;
	showViews = true;

	appState = APP_STATE_NORMAL;

	exePath = NULL;
	char* cmdLine = GetCommandLine();
	int exePathLen = 0;
	if (cmdLine[0] == '\"') {
		cmdLine++;
		while (cmdLine[exePathLen] && (cmdLine[exePathLen] != '\"')) exePathLen++;
	} else {
		while (cmdLine[exePathLen] && (cmdLine[exePathLen] != ' ')) exePathLen++;
	}
	exePath = new char[exePathLen + 1];
	strncpy(exePath, cmdLine, exePathLen);
	do exePath[exePathLen--] = '\0'; while ((exePathLen > 0) && (exePath[exePathLen] != '\\'));
/*
	int currDirLength = GetCurrentDirectory(0, NULL); // length including terminating '\0'
	exePath = new char[currDirLength + 1];
	GetCurrentDirectory(currDirLength, exePath);
	strcat(exePath, "\\");
*/
	readIniFile();

	// init stepper motor class
	stepperMotorUsed = false;
	stepCounter = 0; // for calibration
	smp.setPortAdr(portAdr);

	// init camAccess
	camAccess.init(hwnd, exePath);
}

void ScannerApp::readIniFile() {
	char* ifn = new char[strlen(exePath) + strlen(INI_FILE_NAME) + 1];
	strcpy(ifn, exePath);
	strcat(ifn, INI_FILE_NAME);
	SimpleIni ini(ifn);
	delete[] ifn;

	stepNum360 = 100.0;
	ini.getFloatEntry(INI_STEPNUM360, &stepNum360);

	portAdr = 0;//x0378;
	ini.getIntEntry(INI_PORTADDRESS, &portAdr);

	rotateManually = true;
	int tmp;
	if (ini.getIntEntry(INI_ROTATEMANUALLY, &tmp)) rotateManually = tmp;

	debugOutput = false;
	if (ini.getIntEntry(INI_DEBUGOUTPUT, &tmp)) debugOutput = tmp;

	imageAquisitionType = CamAccess::IAT_TWAIN; // {IAT_FILE, IAT_TWAIN, IAT_VFW}
	char* tmpCharPtr;
	if ((tmpCharPtr=ini.getEntry(INI_IMGAQUISTYPE)) != NULL) {
		if (strcmp(tmpCharPtr, INI_IAT_FILE) == 0) imageAquisitionType = CamAccess::IAT_FILE;
		if (strcmp(tmpCharPtr, INI_IAT_TWAIN) == 0) imageAquisitionType = CamAccess::IAT_TWAIN;
		if (strcmp(tmpCharPtr, INI_IAT_VFW) == 0) imageAquisitionType = CamAccess::IAT_VFW;
	}
	calibrationPattern = CalibrateCam::CAL_PAT_2;
	if ((tmpCharPtr=ini.getEntry(INI_CALIBPATTERN)) != NULL) {
		if (strcmp(tmpCharPtr, INI_CP_1) == 0) calibrationPattern = CalibrateCam::CAL_PAT_1;
		if (strcmp(tmpCharPtr, INI_CP_2) == 0) calibrationPattern = CalibrateCam::CAL_PAT_2;
	}

	// segmentation threshold hack
	if (ini.getIntEntry(INI_SEGLOWTHRESHOLD, &tmp)) Segmentation::THRESHOLD_BG = tmp;
	if (ini.getIntEntry(INI_SEGHIGHTHRESHOLD, &tmp)) Segmentation::THRESHOLD_OBJ = tmp;
}

void ScannerApp::writeIniFile() {
	FILE* iniFile;
	char* ifn = new char[strlen(exePath) + strlen(INI_FILE_NAME) + 1];
	strcpy(ifn, exePath);
	strcat(ifn, INI_FILE_NAME);
	iniFile = fopen(ifn, "w");
	delete[] ifn;
	if (iniFile == NULL) {
		MessageBox(hwnd, "Error while trying to save parameters in ini file", "3DScanner", MB_ICONERROR | MB_OK);
		return;
	}

	fprintf(iniFile, INI_PORTADDRESS "=0x%X\n", portAdr);
	fprintf(iniFile, INI_STEPNUM360 "=%f\n", stepNum360);
	fprintf(iniFile, INI_ROTATEMANUALLY "=%i\n", rotateManually);
	fprintf(iniFile, INI_DEBUGOUTPUT "=%i\n", debugOutput);
	switch (imageAquisitionType) {
		case CamAccess::IAT_FILE:
			fprintf(iniFile, INI_IMGAQUISTYPE "=" INI_IAT_FILE "\n");
			break;
		case CamAccess::IAT_TWAIN:
			fprintf(iniFile, INI_IMGAQUISTYPE "=" INI_IAT_TWAIN "\n");
			break;
		case CamAccess::IAT_VFW:
			fprintf(iniFile, INI_IMGAQUISTYPE "=" INI_IAT_VFW "\n");
			break;
	}
	switch (calibrationPattern) {
		case CalibrateCam::CAL_PAT_1:
			fprintf(iniFile, INI_CALIBPATTERN "=" INI_CP_1 "\n");
			break;
		case CalibrateCam::CAL_PAT_2:
			fprintf(iniFile, INI_CALIBPATTERN "=" INI_CP_2 "\n");
			break;
	}

	// segmentation threshold hack
	fprintf(iniFile, INI_SEGLOWTHRESHOLD "=%i\n", Segmentation::THRESHOLD_BG);
	fprintf(iniFile, INI_SEGHIGHTHRESHOLD "=%i\n", Segmentation::THRESHOLD_OBJ);

	fclose(iniFile);
}

bool ScannerApp::preTranslateMessage(LPMSG msgPtr) {
   	return camAccess.preProcessMessage(msgPtr);
}

void ScannerApp::deinit() {
	if (project) delete project;
	if (mergeProject) delete mergeProject;
	if (stepperMotorUsed) portAccessEnd();
//	if (camAccess) delete camAccess;
	if (exePath) delete[] exePath;
}

////////////////////////////////////////////////////////////////////////////////

void ScannerApp::fileNew() {
	if (project) {
		delete project;
		project = NULL;
	}
	if (mergeProject) {
		delete mergeProject;
		mergeProject = NULL;
	}

	static const int MAX_FILENAME_LENGTH = 500;
	char fileNameBuffer[MAX_FILENAME_LENGTH];
	if (!getFileNameSave(hwnd, (char*) &fileNameBuffer, MAX_FILENAME_LENGTH, (char*) &"3DScanner files\0*.3dscan\0", (char*) &"New 3DScanner project...", "3dscan")) return;

	project = new ScanProject(fileNameBuffer);

	if (!project) {
		MessageBox(hwnd, "Couldn't create project file!", "3DScanner", MB_ICONERROR | MB_OK);
		return;
	}

	appState = APP_STATE_NORMAL;
}

void ScannerApp::fileOpen() {
	static const int MAX_FILENAME_LENGTH = 500;
	char fileNameBuffer[MAX_FILENAME_LENGTH];
	if (!getFileNameOpen(hwnd, (char*) &fileNameBuffer, MAX_FILENAME_LENGTH, (char*) &"3DScanner files\0*.3dscan\0", (char*) &"Open 3DScanner project...", NULL)) return;

	fileOpen(fileNameBuffer);
}

void ScannerApp::fileOpen(char* fileName) {
	if (project) delete project;
	if (mergeProject) {
		delete mergeProject;
		mergeProject = NULL;
	}
	project = new ScanProject(fileName);

	if (!project) {
		MessageBox(hwnd, "Couldn't open project file!", "3DScanner", MB_ICONERROR | MB_OK);
		return;
	}

	if (!project->read()) {
		char buff[50];
		sprintf(buff, "Error while parsing project file in line %i", project->lastErrorLine);
		MessageBox(hwnd, buff, "3DScanner", MB_ICONERROR | MB_OK);
		delete project;
		project = NULL;
	}

	appState = APP_STATE_NORMAL;
}

void ScannerApp::fileSave() {
	if (!project) {
		MessageBox(hwnd, "No open project!", "3DScanner", MB_ICONERROR | MB_OK);
		return;
	}

	project->write();
}

void ScannerApp::fileMergeWithProjectPrepare() {
	// open project to merge with
	static const int MAX_FILENAME_LENGTH = 500;
	char fileNameBuffer[MAX_FILENAME_LENGTH];
	if (!getFileNameOpen(hwnd, (char*) &fileNameBuffer, MAX_FILENAME_LENGTH, (char*) &"3DScanner files\0*.3dscan\0", (char*) &"Open 3DScanner project...", NULL)) return;

	if (mergeProject) delete mergeProject;
	mergeProject = new ScanProject(fileNameBuffer);

	if (!mergeProject) {
		MessageBox(hwnd, "Couldn't open project file!", "3DScanner", MB_ICONERROR | MB_OK);
		return;
	}

	if (!mergeProject->read()) {
		char buff[50];
		sprintf(buff, "Error while parsing project file in line %i", mergeProject->lastErrorLine);
		MessageBox(hwnd, buff, "3DScanner", MB_ICONERROR | MB_OK);
		delete mergeProject;
		mergeProject = NULL;
		return;
	}

	mergeRotation.reset();
	resetVec(&mergeTranslation);
	resetMat(&mergeMatrix);

	// generate 3D Model of second project
	scanGenerate3DModel(mergeProject);

	appState = APP_STATE_MERGE;
}

void ScannerApp::fileMergeWithProjectApply(bool cancel) {
	if (cancel) {
		if (mergeProject) {
			delete mergeProject;
			mergeProject = NULL;
		}

		appState = APP_STATE_NORMAL;
		return;
	}

	/*
	for each photo of mergeProject a new cam def is necessary because
	                                                                       -1                             -1               -1
	mergeMatrix * photo.trafo = mergeMatrix * (camDef.trafo * axisRotation)   = mergeMatrix * axisRotation   * camDef.trafo

	                                                                 -1                                                               -1
	that can't be converted to (new axisRotation without translation)   * (new camDef.trafo independent of rotation angle around axis)
	*/

	// apply rotation and translation
	for (int i=0; i<3; i++) mergeMatrix.m[i][3] *= 1000; // convert translation to millimeters
	for (int i=0; i<mergeProject->photos.size(); i++) { // all photos in mergeProject
		multMatMatInv(&mergeProject->photos[i].camTrafo.trafoInverse, &mergeMatrix);
		getInvRotMat(&mergeProject->photos[i].camTrafo.trafoInverse, &mergeProject->photos[i].camTrafo.trafo);
		mergeProject->photos[i].camTrafo.calcParamsFromTrafoMatrix();
		mergeProject->photos[i].camTrafo.calcTrafoMatrix(); // recalculate trafo matrix to ensure to work with values as they will be when loaded next time
		mergeProject->photos[i].scanParams.camRotationAngle = 0.0; // rotation around axis not used any more
	}

	// copy photo entries and add new camDef for each photo
	for (int i=0; i<mergeProject->photos.size(); i++) { // all photos in mergeProject
		project->photos.push_back(mergeProject->photos[i]);

		// add new cam def with a name that doesn't already exist
		ScanProject::CamDef cdef;
		cdef.camTrafo.copyFrom(&mergeProject->photos[i].camTrafo);
		char* newCdefNameRoot = new char[strlen(mergeProject->camDefs[mergeProject->photos[i].camDefInd].name) + 1];
		char* newCdefName = new char[strlen((char*) &newCdefNameRoot) + 10];
		strcpy((char*) &newCdefNameRoot, mergeProject->camDefs[mergeProject->photos[i].camDefInd].name);
		bool rename;
		int num = 0;
		do {
			char tmpBuffer[9];
			sprintf((char*) &tmpBuffer, " %i", num);
			num++;
			strcpy((char*) newCdefName, (char*) newCdefNameRoot);
			strcat((char*) newCdefName, (char*) &tmpBuffer);
			rename = false;
			for (int j=0; j<project->camDefs.size(); j++) // all camDefs in project
				if (strcmp(newCdefName, project->camDefs[j].name) == 0) rename = true;
		} while (rename);
		// now newCdefName contains a new unused camDef name
		cdef.setName((char*) newCdefName);

		project->camDefs.push_back(cdef);
		project->photos[project->photos.size()-1].camDefInd = project->camDefs.size()-1; // set index to last camDef entry
		project->photos[project->photos.size()-1].calcTrafo(&project->camDefs[0]); // recalculate trafo matrix to ensure to work with values as they will be when loaded next time

		delete[] newCdefNameRoot;
		delete[] newCdefName;
	}

/*
wrong first try
	// apply rotation and translation
	for (int i=0; i<3; i++) mergeMatrix.m[i][3] *= 1000; // convert translation to millimeters
	for (int i=0; i<mergeProject->camDefs.size(); i++) { // all camDefs in mergeProject
		multMatMatInv(&mergeProject->camDefs[i].camTrafo.trafoInverse, &mergeMatrix);
		getInvRotMat(&mergeProject->camDefs[i].camTrafo.trafoInverse, &mergeProject->camDefs[i].camTrafo.trafo);
		mergeProject->camDefs[i].camTrafo.calcParamsFromTrafoMatrix();
		mergeProject->camDefs[i].camTrafo.calcTrafoMatrix(); // recalculate trafo matrix to ensure to work with values as they will be when loaded next time
		multMatVec(&mergeMatrix, &mergeProject->camDefs[i].rotationAxis);
		... wrong ...
	}

	// copy all camdef entries and rename names if necessary
	int oldCamDefsSize = project->camDefs.size();
	for (int i=0; i<mergeProject->camDefs.size(); i++) { // all camDefs in mergeProject
		bool rename = false;

		for (int j=0; j<project->camDefs.size(); j++) // all camDefs in project
			if (strcmp(mergeProject->camDefs[i].name, project->camDefs[j].name) == 0) rename = true;

		if (rename) {
			int num = 0;
			char tmpBuffer[10];
			do {
				rename = false;
				num++;
				sprintf((char*) tmpBuffer, "cam %i", num);
				for (int j=0; j<project->camDefs.size(); j++) // all camDefs in project
					if (strcmp(tmpBuffer, project->camDefs[j].name) == 0) rename = true;
			} while (rename);
			// now tmpBuffer contains a new unused camDef name
			mergeProject->camDefs[i].setName(tmpBuffer);
		}
		project->camDefs.push_back(mergeProject->camDefs[i]);
	}

	// copy and update photo entries
	for (int i=0; i<mergeProject->photos.size(); i++) { // all photos in mergeProject
		project->photos.push_back(mergeProject->photos[i]);
		project->photos[project->photos.size()-1].camDefInd += oldCamDefsSize;
		project->photos[project->photos.size()-1].calcTrafo(&project->camDefs[0]);
	}
*/
	if (mergeProject) {
		delete mergeProject;
		mergeProject = NULL;
	}

	appState = APP_STATE_NORMAL;
	project->modified = true;

	// generate new 3D Model
	scanGenerate3DModel(project);
}

void ScannerApp::fileShowProjectInfo() {
	if (!project) return;
	project->showInfo(hwnd);
}

void ScannerApp::fileExportModel2VRML() { // VRML97
	if (!project || !project->mcPtr) return;

	static const int MAX_FILENAME_LENGTH = 500;
	char fileNameBuffer[MAX_FILENAME_LENGTH];
	if (!getFileNameSave(hwnd, (char*) &fileNameBuffer, MAX_FILENAME_LENGTH, (char*) &"VRML files\0*.wrl\0", (char*) &"Export Model to VRML...", (char*) &"wrl")) return;

	if (!project->mcPtr->write2VRMLFile((char*) &fileNameBuffer, project->isTextured))
		MessageBox(hwnd, "Model couldn't be exported! (File access error?)", "3DScanner", MB_ICONERROR | MB_OK);
}

void ScannerApp::fileExportModel2Obj() {
	if (!project || !project->mcPtr) return;

	static const int MAX_FILENAME_LENGTH = 500;
	char fileNameBuffer[MAX_FILENAME_LENGTH];
	if (!getFileNameSave(hwnd, (char*) &fileNameBuffer, MAX_FILENAME_LENGTH, (char*) &"Wavefront Object files\0*.obj\0", (char*) &"Export Model to OBJ...", (char*) &"obj")) return;

	if (!project->mcPtr->write2ObjFile((char*) &fileNameBuffer))
		MessageBox(hwnd, "Model couldn't be exported! (File access error?)", "3DScanner", MB_ICONERROR | MB_OK);
}

// open PortTalk if not yet opened
bool ScannerApp::portDriverOk() {
	if (stepperMotorUsed) return true;

	if (portAccessBegin(true)) {
		//printf("using PortTalk (running Win NT/2K/ME)\n");
	} else {
		switch (portAccessLastError) {
			case PORTACCES_PORTTALK_NOT_INSTALLED:
				//printf("PortTalk not installed, stopping\n");
				MessageBox(hwnd, "You need to install a driver to grant port access! See menu 'configure'.", "3DScanner", MB_ICONEXCLAMATION | MB_OK);
				return false;
			case PORTACCES_ACCESS_DENIED:
				MessageBox(hwnd, "You don't have the rights to use the port driver, please ask your administrator.", "3DScanner", MB_ICONEXCLAMATION | MB_OK);
				return false;
			case PORTACCES_IOCTL_NOT_IMPLEMENTED:
				//printf("not using PortTalk (running Win9x)\n");
				portAccessBegin(false);
				break;
			case PORTACCES_UNKNOWN_ERROR:
				//printf("unknown error, stopping\n");
				MessageBox(hwnd, "Unknown error during port access driver initialization (porttalk.sys not found in your system directory?)!", "3DScanner", MB_ICONERROR | MB_OK);
				return false;
		}
	}
	stepperMotorUsed = true;
	return true;
}

void ScannerApp::scanRotateTurnTable(int steps) {
	if (rotateManually) {
		MessageBox(hwnd, "rotate now and press OK", "3DScanner", MB_OK);
		return;
	}

	if (!portDriverOk()) return;
	smp.setPower(POWER_HALFSTEP);
	smp.rotateAccelerated(false, steps, 15, 250);
	smp.setPower(POWER_OFF);
	stepCounter += steps; // for calibration
}

// beware of numerical errors when using this repeatedly!
void ScannerApp::rotateTurnTableDeg(int angle) {
	scanRotateTurnTable((int) (stepNum360*angle/360.0));
}

// 1 photo: place calibration object so that the edge that separates the points equals the rotation axis
void ScannerApp::scanCalibrateCam1() {
	if (!project) return;

	MBmpImg* calImg = camAccess.acquireImg(imageAquisitionType, true, "get calibration image...");
	if (!calImg) {
		MessageBox(hwnd, "Image couldn't be aquired!", "3DScanner", MB_ICONERROR | MB_OK);
		return;
	}
	CalibrateCam calCam;
	calCam.setDebugOutput(debugOutput);
	if (!calCam.matchAndCalcParams(calImg, calibrationPattern)) {
		MessageBox(hwnd, "Camera parameters couldn't be calculated!", "3DScanner", MB_ICONERROR | MB_OK);
		delete calImg;
		return;
	}

	calCam.camTrafo.calcTrafoMatrix();

	// add new CamDef
	project->camDefs.push_back(*(new ScanProject::CamDef()));
	MessageBox(hwnd, "New cam definition has been added to project!", "3DScanner", MB_ICONINFORMATION | MB_OK);
	char nbuff[10];
	sprintf(&nbuff[0], "cam %i", project->camDefs.size());
	project->camDefs[project->camDefs.size()-1].setName(&nbuff[0]);

	project->camDefs[project->camDefs.size()-1].camTrafo.copyFrom(&calCam.camTrafo);
	project->camDefs[project->camDefs.size()-1].camTrafo.calcTrafoMatrix();

	// clean up
	delete calImg;
}

// 2 photos
void ScannerApp::scanCalibrateCam() {
	if (!project) return;

	// get two images of calibration object

	if (!portDriverOk()) return;

	// turn turntable
	rotateTurnTableDeg(-8);

	// get first img
//	MBmpImg* calImg1 = camAccess.getSingleImgNoUIModal();
	MBmpImg* calImg1 = camAccess.acquireImg(imageAquisitionType, true, "get first calibration image...");
	if (!calImg1) {
		MessageBox(hwnd, "Image couldn't be aquired!", "3DScanner", MB_ICONERROR | MB_OK);
		return;
	}

	// turn turntable
	rotateTurnTableDeg(+16);

	// get second img
//	MBmpImg* calImg2 = camAccess.getSingleImgNoUIModal();
	MBmpImg* calImg2 = camAccess.acquireImg(imageAquisitionType, true, "get second calibration image...");
	if (!calImg2) {
		MessageBox(hwnd, "Image couldn't be aquired!", "3DScanner", MB_ICONERROR | MB_OK);
		delete calImg1;
		return;
	}


//MBmpImg* calImg1 = new MBmpImg("..\\cam calibration\\drehachse1_.bmp");
//MBmpImg* calImg2 = new MBmpImg("..\\cam calibration\\drehachse2_.bmp");

	if ((calImg1->getWidth() < 50) || (calImg2->getWidth() < 50)) {
		MessageBox(hwnd, "Calibration photos invalid.", "3DScanner", MB_ICONERROR | MB_OK);
		delete calImg1;
		delete calImg2;
		return;
	}

	setupProgressBar(hwnd, hInstance, 0, 3);
	incrementProgressBar();

	// calculate calibration parameters of both views
	CalibrateCam calCam1, calCam2;
	calCam1.setDebugOutput(debugOutput);
	calCam2.setDebugOutput(debugOutput);

	bool macp1 = calCam1.matchAndCalcParams(calImg1, calibrationPattern);
	incrementProgressBar();
	bool macp2; if (macp1) macp2 = calCam2.matchAndCalcParams(calImg2, calibrationPattern);

	if (!macp1 || !macp2) {
		MessageBox(hwnd, "Camera parameters couldn't be calculated!", "3DScanner", MB_ICONERROR | MB_OK);
		closeProgressBar();
		delete calImg1;
		delete calImg2;
		return;
	}

	if (debugOutput) {
		calCam1.writeMatchToFile("calCam1Match.txt");
		calCam1.writeParamsToFile("calCam1Params.txt");
		calCam2.writeMatchToFile("calCam2Match.txt");
		calCam2.writeParamsToFile("calCam2Params.txt");
	}

	incrementProgressBar();

	calCam1.camTrafo.calcTrafoMatrix();
	calCam2.camTrafo.calcTrafoMatrix();

/*
	MAT4x4 w1ToW2;
	copyMat(&calCam2.camTrafo.trafo, &w1ToW2);
	multMatMat(&w1ToW2, &calCam1.camTrafo.trafoInverse);
	... mit   w1ToW2 * d = d
	     ==> (W1ToW2 - I) * d = 0   kann man das lsen (Form Ax=b)
*/

/* TODO (#1#): estimate rotation axis, too
               (and edit camDef.rotationAxis) */

	// determine axis by intersecting perpendicular lines to same points of both images

	struct line2dStruct {double x1, y1, x2, y2;};
	double tmpX, tmpY, tmpZ, tmp;

	line2dStruct linesConn[3];
	linesConn[0].x1 = 100;
	linesConn[0].y1 = 0;
	calCam1.camTrafo.world2Cam(linesConn[0].x1, linesConn[0].y1, 0, &tmpX, &tmpY, &tmpZ);
	calCam2.camTrafo.cam2World(tmpX, tmpY, tmpZ, &linesConn[0].x2, &linesConn[0].y2, &tmp);

	linesConn[1].x1 = 0;
	linesConn[1].y1 = 100;
	calCam1.camTrafo.world2Cam(linesConn[1].x1, linesConn[1].y1, 0, &tmpX, &tmpY, &tmpZ);
	calCam2.camTrafo.cam2World(tmpX, tmpY, tmpZ, &linesConn[1].x2, &linesConn[1].y2, &tmp);

	if (debugOutput) {
		linesConn[2].x1 = 0;
		linesConn[2].y1 = 0;
		calCam1.camTrafo.world2Cam(linesConn[2].x1, linesConn[2].y1, 0, &tmpX, &tmpY, &tmpZ);
		calCam2.camTrafo.cam2World(tmpX, tmpY, tmpZ, &linesConn[2].x2, &linesConn[2].y2, &tmp);

		for (int i = 0; i<2; i++) {
			MBmpImg* imgPtr = (i) ? calImg2 : calImg1;
			double imX, imY, imX2, imY2;

			calCam2.camTrafo.world2Img(linesConn[2].x1, linesConn[2].y1, 0, &imX, &imY);
			calCam2.camTrafo.world2Img(linesConn[0].x1, linesConn[0].y1, 0, &imX2, &imY2);
			imgPtr->lineRGB((int) imX, (int) imY, (int) imX2, (int) imY2, 0, 0, 255);

			calCam2.camTrafo.world2Img(linesConn[2].x1, linesConn[2].y1, 0, &imX, &imY);
			calCam2.camTrafo.world2Img(linesConn[1].x1, linesConn[1].y1, 0, &imX2, &imY2);
			imgPtr->lineRGB((int) imX, (int) imY, (int) imX2, (int) imY2, 0, 0, 255);

			calCam2.camTrafo.world2Img(linesConn[2].x2, linesConn[2].y2, 0, &imX, &imY);
			calCam2.camTrafo.world2Img(linesConn[0].x2, linesConn[0].y2, 0, &imX2, &imY2);
			imgPtr->lineRGB((int) imX, (int) imY, (int) imX2, (int) imY2, 255, 0, 0);

			calCam2.camTrafo.world2Img(linesConn[2].x2, linesConn[2].y2, 0, &imX, &imY);
			calCam2.camTrafo.world2Img(linesConn[1].x2, linesConn[1].y2, 0, &imX2, &imY2);
			imgPtr->lineRGB((int) imX, (int) imY, (int) imX2, (int) imY2, 255, 0, 0);

			calCam2.camTrafo.world2Img(linesConn[0].x1, linesConn[0].y1, 0, &imX, &imY);
			BYTE* bPtr = imgPtr->getDataXY((int) imX, (int) imY);
			*bPtr++ = 255;
			*bPtr++ = 0;
			*bPtr = 0;

			calCam2.camTrafo.world2Img(linesConn[0].x2, linesConn[0].y2, 0, &imX, &imY);
			bPtr = imgPtr->getDataXY((int) imX, (int) imY);
			*bPtr++ = 255;
			*bPtr++ = 100;
			*bPtr = 100;

			calCam2.camTrafo.world2Img(linesConn[1].x1, linesConn[1].y1, 0, &imX, &imY);
			bPtr = imgPtr->getDataXY((int) imX, (int) imY);
			*bPtr++ = 0;
			*bPtr++ = 255;
			*bPtr = 0;

			calCam2.camTrafo.world2Img(linesConn[1].x2, linesConn[1].y2, 0, &imX, &imY);
			bPtr = imgPtr->getDataXY((int) imX, (int) imY);
			*bPtr++ = 100;
			*bPtr++ = 255;
			*bPtr = 100;

			calCam2.camTrafo.world2Img(linesConn[2].x1, linesConn[2].y1, 0, &imX, &imY);
			bPtr = imgPtr->getDataXY((int) imX, (int) imY);
			*bPtr++ = 0;
			*bPtr++ = 0;
			*bPtr = 255;

			calCam2.camTrafo.world2Img(linesConn[2].x2, linesConn[2].y2, 0, &imX, &imY);
			bPtr = imgPtr->getDataXY((int) imX, (int) imY);
			*bPtr++ = 100;
			*bPtr++ = 100;
			*bPtr = 255;
		}
	}

	line2dStruct lp[2]; // perpendicular lines
	for (int i=0; i<2; i++) {
		lp[i].x1 = 0.5 * (linesConn[i].x1 + linesConn[i].x2);
		lp[i].y1 = 0.5 * (linesConn[i].y1 + linesConn[i].y2);
		lp[i].x2 = linesConn[i].y2 - linesConn[i].y1;
		lp[i].y2 = linesConn[i].x1 - linesConn[i].x2;
	}

	// lp[0].1 + a*lp[0].2 = lp[1].1 + b*lp[1].2
	float b1 = lp[0].x2*(lp[1].y1-lp[0].y1) - lp[0].y2*(lp[1].x1-lp[0].x1);
	float b2 = lp[0].y2*lp[1].x2 - lp[0].x2*lp[1].y2;
	if (b2 == 0) {
		MessageBox(hwnd, "Camera parameters couldn't be calculated!", "3DScanner", MB_ICONERROR | MB_OK);
		closeProgressBar();
		delete calImg1;
		delete calImg2;
		return;
	}
	float b = b1 / b2;

	float axisX = lp[1].x1 + b*lp[1].x2;
	float axisY = lp[1].y1 + b*lp[1].y2;

	calCam2.camTrafo.tx += calCam2.camTrafo.trafo.m[0][0] * axisX + calCam2.camTrafo.trafo.m[0][1] * axisY;
	calCam2.camTrafo.ty += calCam2.camTrafo.trafo.m[1][0] * axisX + calCam2.camTrafo.trafo.m[1][1] * axisY;
	calCam2.camTrafo.tz += calCam2.camTrafo.trafo.m[2][0] * axisX + calCam2.camTrafo.trafo.m[2][1] * axisY;

	calCam2.camTrafo.calcTrafoMatrix();

	if (debugOutput) {
		double imX, imY, imX2, imY2;

		calCam2.camTrafo.world2Img(0, 0, 0, &imX, &imY);
		/*
		calCam2.camTrafo.world2Img(100, 0, 0, &imX, &imY);
		sprintf((char*) &buf, "x: %f, y: %f", imX, imY);
		MessageBox(hwnd, buf, "100", MB_OK);

		calCam2.camTrafo.world2Img(0, 100, 0, &imX, &imY);
		sprintf((char*) &buf, "x: %f, y: %f", imX, imY);
		MessageBox(hwnd, buf, "010", MB_OK);
		*/
		calCam2.camTrafo.world2Img(0, 0, 100, &imX2, &imY2);
		calImg1->lineRGB((int) imX, (int) imY, (int) imX2, (int) imY2, 255, 0, 255);
		calImg2->lineRGB((int) imX, (int) imY, (int) imX2, (int) imY2, 255, 0, 255);

		calImg1->write2BmpFile("_drehachse2.bmp");
		calImg2->write2BmpFile("_drehachse.bmp");
	}

	// add new CamDef
	project->camDefs.push_back(*(new ScanProject::CamDef()));
	MessageBox(hwnd, "New cam definition has been added to project!", "3DScanner", MB_ICONINFORMATION | MB_OK);
	char nbuff[10];
	sprintf(&nbuff[0], "cam %i", project->camDefs.size());
	project->camDefs[project->camDefs.size()-1].setName(&nbuff[0]);

	project->camDefs[project->camDefs.size()-1].camTrafo.copyFrom(&calCam2.camTrafo);
	project->camDefs[project->camDefs.size()-1].camTrafo.calcTrafoMatrix();

	// clean up
	closeProgressBar();
	delete calImg1;
	delete calImg2;
}

void ScannerApp::scanGetImages() { // int param
	if (!project) return;

	if (project->camDefs.size() == 0) {
		MessageBox(hwnd, "Calibrate the camera first!", "3DScanner", MB_ICONERROR | MB_OK);
		return;
	}

	// exists already a photo with this cam def?
	bool phExists = false;
	for (int i=0; i<project->photos.size(); i++) {
		if (project->photos[i].camDefInd == project->camDefs.size()-1) phExists = true;
	}
	if (phExists) {
		MLInputDialog bInpDlg(hwnd, hInstance);
		BOOL answ;
		if (!bInpDlg.getBool((char*) &"Overwrite existant photos with this camera definition?", &answ)) return; // return if canceled
		if (!answ) return;

		// yes: overwrite; here: delete them
		int i=0;
		while (i<project->photos.size()) {
			if (project->photos[i].camDefInd == project->camDefs.size()-1) {
// ???				project->photos.erase(i);
				ScanProject::Photo tmpPh = project->photos[project->photos.size()-1];
				project->photos[project->photos.size()-1] = project->photos[i];
				project->photos[i] = tmpPh;
				project->photos.pop_back();
			} else i++;
		}
	}

	MLInputDialog inpDlg(hwnd, hInstance);
	long int phNumLong;
	if (!inpDlg.getLongMinMax((char*) &"Number of photos to take with this camera position", (char*) &"8", 1, 36, &phNumLong)) return; // return if canceled

	setupProgressBar(hwnd, hInstance, 0, phNumLong);
	for (int photosTaken=0; photosTaken<phNumLong; photosTaken++) {
		// spin turntable
		int lastStep = (int) stepNum360*photosTaken/phNumLong;
		int nowStep = (int) stepNum360*(photosTaken+1)/phNumLong;
		if (phNumLong > 1) scanRotateTurnTable(nowStep - lastStep);
//		smp.setPower(POWER_HALFSTEP);
//		smp.rotateAccelerated(false, nowStep - lastStep, 15, 250);
//		smp.setPower(POWER_OFF);

		// get image and save to file
		char* imgFileName = new char[strlen(project->projectName) + 10];
		sprintf(imgFileName, "%s %i.bmp", project->projectName, photosTaken+1);
		if (!camAccess.acquireImgAndSave2File(imageAquisitionType, false, "get image...", imgFileName)) {
			MessageBox(hwnd, "warning: scan incomplete\nnot all images could be aquired", "3DScanner", MB_ICONERROR | MB_OK);
			closeProgressBar();
			return;
		}
		incrementProgressBar();

		project->modified = true;
		ScanProject::Photo photo;
		photo.camDefInd = project->camDefs.size()-1;
		photo.setFileName(imgFileName);
		photo.scanParams.camRotationAngle = -360.0 * (photosTaken+1)/phNumLong;
//		photo.camTrafo.calcTrafoMatrix();
		photo.calcTrafo(&project->camDefs[photo.camDefInd]);
		project->photos.push_back(photo);
		delete[] imgFileName;
	}

	closeProgressBar();

/*
	switch (param) {
		case GET_IMG_START: {
			if (!portDriverOk()) return;
			if (!camAccess.initOk(hwnd)) return;
/ *			if (!camAccess.prepareGetMultiImgNoUI()) {
				MessageBox(hwnd, "Twain Source could not be opened!", "3DScanner", MB_ICONERROR | MB_OK);
				return;
			}
* /
/ *
if (project->camDefs.size() == 0) {
// add camDef
project->camDefs.push_back(*(new ScanProject::CamDef()));
MessageBox(hwnd, "New cam definition has been added to project!", "3DScanner", MB_ICONERROR | MB_OK);
char nbuff[10];
sprintf(&nbuff[0], "cam %i", project->camDefs.size());
project->camDefs[project->camDefs.size()-1].setName(&nbuff[0]);
}
* /
			if (project->camDefs.size() == 0) {
				MessageBox(hwnd, "Calibrate the camera first!", "3DScanner", MB_ICONERROR | MB_OK);
				return;
			}


			//photoNum = 8;

			MLInputDialog inpDlg(hwnd, hInstance);
			long int phNumLong;
			if (!inpDlg.getLongMinMax((char*) &"Number of photos to take with this camera position", (char*) &"8", 1, 36, &phNumLong)) return; // return if canceled
			project->photoNum = phNumLong;

			project->photosTaken = 0;
			//PostMessage(hwnd, WM_USER + MSG_GET_IMG, GET_IMG_CONTINUE, 0);
			//break;
			}
			// fall through
		case GET_IMG_CONTINUE:
			if (project->photoNum == 0) return;
			if (project->photosTaken == project->photoNum) {
//				camAccess.stopGetMultiImgNoUI();
				return;
			}

			// spin turntable
			int lastStep = (int) stepNum360*project->photosTaken/project->photoNum;
			int nowStep = (int) stepNum360*(project->photosTaken+1)/project->photoNum;
			smp.setPower(POWER_HALFSTEP);
			smp.rotateAccelerated(false, nowStep - lastStep, 15, 250);
			smp.setPower(POWER_OFF);

			// get image via TWAIN and save to file
			char imgFileName[100];
			sprintf(imgFileName, "test %i.bmp", project->photosTaken+1);
//			camAccess.acquireImgNoUIAndSave2File(imgFileName);
			if (!camAccess.acquireImgAndSave2File(imageAquisitionType, false, "get image...", imgFileName)) {
				MessageBox(hwnd, "3DScanner", "warning: scan incomplete\nnot all images could be aquired", MB_ICONERROR | MB_OK);
				return;
			}

			project->modified = true;
			ScanProject::Photo photo;
			photo.camDefInd = project->camDefs.size()-1;
			photo.setFileName(imgFileName);
			photo.scanParams.camRotationAngle = -360.0 * (project->photosTaken+1)/project->photoNum;
			photo.camTrafo.calcTrafoMatrix();
			project->photos.push_back(photo);

			project->photosTaken++; // photo is actually not taken yet but nearly
			break;
	}
*/
}

void ScannerApp::calcMaxScanningVolume() {
	// find roughly minimum and maximum dimensions for scanning volume:
	// for each photo calc max radius and max z of closest point on a camera corner ray to z axis
	// (except if camera looking from above: then take cam position as z and max radius on xy plane)
	// take minimum radius and z of photos

	double ax, ay, az, bx, by, bz;
	float minRadA = 1000, minZA = 1000;
	for (int i=0; i<project->photos.size(); i++) {
		int iw = project->photos[i].camTrafo.imgWidth - 1;
		int ih = project->photos[i].camTrafo.imgHeight - 1;
		float maxRadP = 0, maxZP = 0;
		for (int x=0; x<2; x++) for (int y=0; y<2; y++) {
			project->photos[i].camTrafo.img2World(x*iw, y*ih, &ax, &ay, &az, &bx, &by, &bz);
			VEC4 a, b, p;
			resetVec(&a);
			resetVec(&b);
			resetVec(&p);
			a.v[0] = ax * 0.001; // a: cam position
			a.v[1] = ay * 0.001;
			a.v[2] = az * 0.001;
			b.v[0] = bx;
			b.v[1] = by;
			b.v[2] = bz;
			normalizeVec3(&b);
			// now a + m*b is ray pointing into the scene
			if (fabs(b.v[2] > 0.75)) {
				// bird's view (looking down or up)
				// calc m for intersection with xy plane
				float mXYPlane = -a.v[2] / b.v[2];
				addVec(&p, &a); // p = a + mXYPlane*b
				scaleVec(&b, mXYPlane);
				addVec(&p, &b);

				p.v[2] = fabs(a.v[2]); // camera z position
			} else {
				// calc m for nearest point to z axis
				float mNearZ = (-a.v[1]*b.v[1] - a.v[0]*b.v[0]) / (sqr(b.v[0]) + sqr(b.v[1]));
				addVec(&p, &a); // p = a + mNearZ*b
				scaleVec(&b, mNearZ);
				addVec(&p, &b);
			}
			float rad = sqrt(sqr(p.v[0]) + sqr(p.v[1]));

			if (p.v[2] > maxZP) maxZP = p.v[2];
			if (rad > maxRadP) maxRadP = rad;
		}
		if (maxZP < minZA) minZA = maxZP;
		if (maxRadP < minRadA) minRadA = maxRadP;
	}
/*
char buff[200];
sprintf(buff, "%f %f", minRadA, minZA);
MessageBox(NULL, buff, buff, MB_OK);
*/

	project->minX = -minRadA;
	project->maxX = minRadA;
	project->minY = -minRadA;
	project->maxY = minRadA;
	project->minZ = -minZA * 0.01;
	project->maxZ = minZA;
/*
MessageBox(hwnd, "max scanning volume calculated", "3DScanner", MB_OK);
if (project->minX != 0) MessageBox(hwnd, "max scanning volume ok", "3DScanner", MB_OK);
*/
}

void ScannerApp::scanSegmentImages() {
	if (!project) return;
	if (!project->photos.size()) return;

	setupProgressBar(hwnd, hInstance, 0, project->photos.size()-1);
	for (int i=0; i<project->photos.size(); i++)  // all photos
			if (project->photos[i].fileName) {
			// open rgb image
			MBmpImg photo(project->photos[i].fileName);
			MBmpImg mask(photo.getWidth(), photo.getHeight(), 8);
			Segmentation::segment(&photo, &mask);
/*
			int width = photo.getWidth();
			int height = photo.getHeight();
			// create image that will containt median filtered photo
			MBmpImg photoMedian(width, height, 24);

			// create alpha image
			MBmpImg mask1(width, height, 8);
			MBmpImg mask2(width, height, 8);

			photo.tmMedianRGB(&photoMedian, 2.0);
photoMedian.write2BmpFile("segmtest_median.bmp");

			photoMedian.sobelRGB(&photo, 5.0);
photo.write2BmpFile("segmtest_median_sobel.bmp");


			// estimate background color: take color near the upper left corner
			BYTE* bgDatPtr = photoMedian.getDataXY((width>50)?10:0, (height>50)?10:0);
			int bgR = *bgDatPtr++;
			int bgG = *bgDatPtr++;
			int bgB = *bgDatPtr;

			int bgH, bgS, bgL;
			MBmpImg::getHSL(bgR, bgG, bgB, &bgH, &bgS, &bgL);

			// generate color difference image to background color (HSL model)
			for (int y=0; y<height; y++)
				for (int x=0; x<width; x++) { // all pixel
					BYTE* datPtr = photoMedian.getDataXY(x, y);
					BYTE r = *datPtr++;
					BYTE g = *datPtr++;
					BYTE b = *datPtr;

					int h, s, l;
					MBmpImg::getHSL(r, g, b, &h, &s, &l);

					int dl = abs(bgL - l) - 30; if (dl < 0) dl = 0; dl *= 5; //dl = sqr(dl);
					int ds = abs(bgS - s) - 50; if (ds < 0) ds = 0; ds = sqr(ds);
					int dh = abs(bgH - h);
					if (dh>180) dh = 360-dh;
					dh -= 5; if (dh < 0) dh = 0; dh *= 40;// dh = sqr(dh);

					int val = dl + ds + dh;

					// int val = sqr(bgR - r)/10 + sqr(bgG - g)/10 + sqr(bgB - b)/10;
					if (val < 0) val = 0;
					if (val > 255) val = 255;
					*mask1.getDataXY(x, y) = val;
				}

mask1.write2BmpFile("segmtestm no rg.bmp");

			// region growing of background (black) with edge detected border
/ * better: while (new pixels are drawn) anstatt for (1 to 5) * /
for (int rg=0; rg<5; rg++) {
			for (int y=1; y<height; y++)
			 for (int x=1; x<width; x++) { // all pixel except 1 pixel border on left and top

				int m = *mask1.getDataXY(x, y);
				BYTE ed = (((int) *photo.getDataXY(x, y)) + *(photo.getDataXY(x, y)+1) + *(photo.getDataXY(x, y)+2))/3;

				if ((m > 0) && (ed < 50)) {
					BYTE nc = *mask1.getDataXY(x-1, y);
					if (nc +5< m) m = nc+5;
					nc = *mask1.getDataXY(x, y-1);
					if (nc +5< m) m = nc+5;
				}
if (m<0) m=0;
if (m>255) m=255;
				*mask1.getDataXY(x, y) = m;

			}

			for (int y=height-2; y>=0; y--)
			 for (int x=width-2; x>=0; x--) { // all pixel except 1 pixel border on right and bottom
				int m = *mask1.getDataXY(x, y);
				BYTE ed = (((int) *photo.getDataXY(x, y)) + *(photo.getDataXY(x, y)+1) + *(photo.getDataXY(x, y)+2))/3;

				if ((m > 0) && (ed < 50)) {
					BYTE nc = *mask1.getDataXY(x+1, y);
					if (nc +5< m) m = nc+5;
					nc = *mask1.getDataXY(x, y+1);
					if (nc +5< m) m = nc+5;
				}
if (m<0) m=0;
if (m>255) m=255;
				*mask1.getDataXY(x, y) = m;
			}

}
mask1.write2BmpFile("segmtestm with rg.bmp");


			// opening with 3x3
			mask1.erodeGray(&mask2, 3);
			mask2.dilateGray(&mask1, 3);
mask1.write2BmpFile("segmtestm opened.bmp");

			// blur mask
			mask1.blurGaussian121(&mask2);
mask1.write2BmpFile("segmtestm blured.bmp");

mask1.write2BmpFile("segmtestm full.bmp");
*/
			// save alpha image
			if (project->photos[i].maskFileName) delete project->photos[i].maskFileName;
			project->photos[i].maskFileName = new char[strlen(project->photos[i].fileName)+2];
			strcpy(project->photos[i].maskFileName, project->photos[i].fileName);
			char* maskEnd = project->photos[i].maskFileName;
			maskEnd += strlen(project->photos[i].fileName)-4;
			strcpy(maskEnd, "m.bmp");

			mask.write2BmpFile(project->photos[i].maskFileName);

			//MessageBox(hwnd, project->photos[i].maskFileName, "saved", MB_OK);

			project->modified = true;
//			project->photos[i].setMaskFileName(project->photos[i].maskFileName);

			incrementProgressBar();
		}
		closeProgressBar();
}

void ScannerApp::scanPrepareManualSegmentation() {
	if (!project) return;
	if (!project->photos.size()) return;

	setupProgressBar(hwnd, hInstance, 0, project->photos.size()-1);
	for (int i=0; i<project->photos.size(); i++)  // all photos
			if (project->photos[i].fileName) {
			// open rgb image
			MBmpImg photo(project->photos[i].fileName);
			int width = photo.getWidth();
			int height = photo.getHeight();
			MBmpImg photoHSL(width, height, 24);
			photo.RGB2HSL(&photoHSL);
			MBmpImg mask(width, height, 8);

			// copy luminance to mask image
			for (int y=0; y<height; y++)
				for (int x=0; x<width; x++) { // all pixel
					BYTE* datPtr = photoHSL.getDataXY(x, y);
					BYTE h = *datPtr++;
					BYTE s = *datPtr++;
					BYTE l = *datPtr;
					*mask.getDataXY(x, y) = l;
				}

			// save alpha image
			if (project->photos[i].maskFileName) delete project->photos[i].maskFileName;
			project->photos[i].maskFileName = new char[strlen(project->photos[i].fileName)+2];
			strcpy(project->photos[i].maskFileName, project->photos[i].fileName);
			char* maskEnd = project->photos[i].maskFileName;
			maskEnd += strlen(project->photos[i].fileName)-4;
			strcpy(maskEnd, "m.bmp");

			mask.write2BmpFile(project->photos[i].maskFileName);

			//MessageBox(hwnd, project->photos[i].maskFileName, "saved", MB_OK);

			project->modified = true;
//			project->photos[i].setMaskFileName(project->photos[i].maskFileName);

			incrementProgressBar();
		}
		closeProgressBar();
}

void ScannerApp::scanGenerate3DModel(ScanProject* prj) {
	if (!prj) return;
	if (!prj->photos.size()) return;

	// calc max scanning volume if not yet done
	if ((prj->maxX - prj->minX == 0) || (prj->maxY - prj->minY == 0)
		|| (prj->maxZ - prj->minZ == 0))
		calcMaxScanningVolume();

char* LOG_FILE_NAME = "log.txt";
char* lfn = new char[strlen(exePath) + strlen(LOG_FILE_NAME) + 1];
strcpy(lfn, exePath);
strcat(lfn, LOG_FILE_NAME);

FILE* logFile;
logFile = fopen(lfn, "w");
delete[] lfn;
fprintf(logFile, "scanGenerate3DModel\n\n");
/* TODO (#1#): better would be a global logfile for this whole class! */

	// open all alpha images
	MultiResImgPtr* img = new MultiResImgPtr[prj->photos.size()];
	bool allImgsLoaded = true;
	for (int i=0; i<prj->photos.size(); i++) {
fprintf(logFile, "opening %s\n", prj->photos[i].maskFileName);
		img[i] = new MultiResImg(prj->photos[i].maskFileName);
		if (img[i]->getWidth() == 0) allImgsLoaded = false;
		if (img[i]->getBitsPerPixel() != 8) allImgsLoaded = false;
	}

	if (!allImgsLoaded) {
		MessageBox(hwnd, "Not all images could be loaded (use 8 bit grayscale bitmaps)!", "3DScanner", MB_ICONERROR | MB_OK);
		// clean up
		for (int i=0; i<prj->photos.size(); i++) delete img[i];
		delete[] img;
		return;
	}

	//////////////////////////////////////////////////////////////////
	////  generate triangles with marching cubes

	// edge length of each cube
	MLInputDialog resInpDlg(hwnd, hInstance);
	double resDouble;
	float minDim = prj->maxX - prj->minX;
	if (prj->maxY - prj->minY < minDim) minDim = prj->maxY - prj->minY;
	if (prj->maxZ - prj->minZ < minDim) minDim = prj->maxZ - prj->minZ;
	float maxDim = prj->maxX - prj->minX;
	if (prj->maxY - prj->minY > maxDim) maxDim = prj->maxY - prj->minY;
	if (prj->maxZ - prj->minZ > maxDim) maxDim = prj->maxZ - prj->minZ;
	char defValString[20];
	float resDefault = 50/maxDim;
	float resMin = 5/minDim;
	float resMax = 1000/maxDim;
	if (resDefault < resMin) resDefault = resMin;
	if (resDefault > resMax) resDefault = resMax;
	sprintf((char*) &defValString, "%f", resDefault);
	if (!resInpDlg.getDoubleMinMax((char*) &"Resolution: number of cubes per meter", (char*) &defValString, resMin, resMax, &resDouble)) return; // return if canceled
	prj->resolution = 1/resDouble;

	// number of cubes in each dimension
	prj->gridSizeX = (int) ceil((prj->maxX - prj->minX) / prj->resolution + 1);
	prj->gridSizeY = (int) ceil((prj->maxY - prj->minY) / prj->resolution + 1);
	prj->gridSizeZ = (int) ceil((prj->maxZ - prj->minZ) / prj->resolution + 1);
	if (prj->mcPtr) delete prj->mcPtr;
	prj->mcPtr = new MarchingCubesVertexExt();
	prj->isTextured = false;

	char *slices[2];
	slices[0] = new char[prj->gridSizeY * prj->gridSizeX];
	slices[1] = new char[prj->gridSizeY * prj->gridSizeX];

/* fr Koordinatenanzeige im Kalibrierungsbild (--> stimmt alles?)
{
double imX, imY;
char buf[100];

prj->photos[0].camTrafo.world2Img(0, 0, 0, &imX, &imY);
sprintf((char*) &buf, "x: %f, y: %f", imX, imY);
MessageBox(hwnd, buf, "000", MB_OK);

prj->photos[0].camTrafo.world2Img(100, 0, 0, &imX, &imY);
sprintf((char*) &buf, "x: %f, y: %f", imX, imY);
MessageBox(hwnd, buf, "100", MB_OK);

prj->photos[0].camTrafo.world2Img(0, 100, 0, &imX, &imY);
sprintf((char*) &buf, "x: %f, y: %f", imX, imY);
MessageBox(hwnd, buf, "010", MB_OK);

prj->photos[0].camTrafo.world2Img(0, 0, 100, &imX, &imY);
sprintf((char*) &buf, "x: %f, y: %f", imX, imY);
MessageBox(hwnd, buf, "001", MB_OK);
}
*/

	fprintf(logFile, "mc.init():\n");
	prj->mcPtr->init(prj->gridSizeX, prj->gridSizeY, prj->gridSizeZ,
		0.5*(prj->gridSizeX-1)*prj->resolution, 0.5*(prj->gridSizeY-1)*prj->resolution,
		0.5*(prj->gridSizeZ-1)*prj->resolution);
	prj->mcPtr->translate(-(prj->maxX + prj->minX)/2, -(prj->maxY + prj->minY)/2, -(prj->gridSizeZ-1 - prj->CLIP_XY_PLANE_ZSUB) * 0.5);

prj->mcPtr->translX = -prj->minX / prj->resolution;
prj->mcPtr->translY = -prj->minY / prj->resolution;
prj->mcPtr->translZ = -prj->minZ / prj->resolution - prj->CLIP_XY_PLANE_ZSUB;

	setupProgressBar(hwnd, hInstance, 0, prj->gridSizeZ-1);

	lastScanMinX = (prj->maxX + prj->minX) / 2;
	lastScanMaxX = lastScanMinX;
	lastScanMinY = (prj->maxY + prj->minY) / 2;
	lastScanMaxY = lastScanMinY;
	lastScanMinZ = (prj->maxZ + prj->minZ) / 2;
	lastScanMaxZ = lastScanMinZ;

	fprintf(logFile, "filling data array:\n");
	for (int z=0; z<prj->gridSizeZ; z++) {
		// fill data array
		for (int y=0; y<prj->gridSizeY; y++) {
			for (int x=0; x<prj->gridSizeX; x++) {
				int val = 255;
				// calc coords in world coordinates
/*				float wx = prj->resolution*(x - 0.5*(prj->gridSizeX-1));
				float wy = prj->resolution*(y - 0.5*(prj->gridSizeY-1));
				float wz = prj->resolution*(z - prj->CLIP_XY_PLANE_ZSUB);
*/
				float wx = (x - prj->mcPtr->translX) * prj->mcPtr->tScaleX;
				float wy = (y - prj->mcPtr->translY) * prj->mcPtr->tScaleY;
				float wz = (z - prj->mcPtr->translZ) * prj->mcPtr->tScaleZ;

				for (int i=0; i<prj->photos.size(); i++) { // all images
					double imX, imY;
					prj->photos[i].camTrafo.world2Img(wx*1000, wy*1000, wz*1000, &imX, &imY);

/*
					// make cube corners to object points
					if (((x<2) || (x>prj->gridSizeX-3)) && ((y<2) || (y>prj->gridSizeY-3)) && ((z<2) || (z>prj->gridSizeZ-3))) {
						val = 255;
						break;
					}
*/
					// skip if outside image boundaries
					if ((imX >= 0) && (imX < img[i]->getWidth())
						&& (imY >= 0) && (imY < img[i]->getHeight())) {

						// calculation of influence Diameter (for anti aliasing)
						// add distances of image coords for next grid position in each dimension
						// _SLOW_ !!!
						float inflDiam = 0;
						double imTX, imTY;
/*						float wx = prj->resolution*(x + 1 - 0.5*(prj->gridSizeX-1));
						float wy = prj->resolution*(y - 0.5*(prj->gridSizeY-1));
						float wz = prj->resolution*(z - prj->CLIP_XY_PLANE_ZSUB);
*/
				float wx = (x + 1 - prj->mcPtr->translX) * prj->mcPtr->tScaleX;
				float wy = (y - prj->mcPtr->translY) * prj->mcPtr->tScaleY;
				float wz = (z - prj->mcPtr->translZ) * prj->mcPtr->tScaleZ;
						prj->photos[i].camTrafo.world2Img(wx*1000, wy*1000, wz*1000, &imTX, &imTY);

						inflDiam += sqrt(sqr(imX - imTX) + sqr(imY - imTY));

/*						wx = prj->resolution*(x - 0.5*(prj->gridSizeX-1));
						wy = prj->resolution*(y + 1 - 0.5*(prj->gridSizeY-1));
						wz = prj->resolution*(z - prj->CLIP_XY_PLANE_ZSUB);
*/
				wx = (x - prj->mcPtr->translX) * prj->mcPtr->tScaleX;
				wy = (y + 1 - prj->mcPtr->translY) * prj->mcPtr->tScaleY;
				wz = (z - prj->mcPtr->translZ) * prj->mcPtr->tScaleZ;
						prj->photos[i].camTrafo.world2Img(wx*1000, wy*1000, wz*1000, &imTX, &imTY);
						inflDiam += sqrt(sqr(imX - imTX) + sqr(imY - imTY));

/*						wx = prj->resolution*(x - 0.5*(prj->gridSizeX-1));
						wy = prj->resolution*(y - 0.5*(prj->gridSizeY-1));
						wz = prj->resolution*(z + 1 - prj->CLIP_XY_PLANE_ZSUB);
*/
				wx = (x - prj->mcPtr->translX) * prj->mcPtr->tScaleX;
				wy = (y - prj->mcPtr->translY) * prj->mcPtr->tScaleY;
				wz = (z + 1 - prj->mcPtr->translZ) * prj->mcPtr->tScaleZ;
						prj->photos[i].camTrafo.world2Img(wx*1000, wy*1000, wz*1000, &imTX, &imTY);
						inflDiam += sqrt(sqr(imX - imTX) + sqr(imY - imTY));

						inflDiam *= 0.5;

						int imgVal = img[i]->getGrayTriLinIntPolXY(imX, imY, inflDiam);

						//int imgVal = ((abs(imX - camParams[i].ox) < 100) && (abs(imY - camParams[i].oy) < 70))?255:0; // test for intersecting volume

						if (imgVal < val) val = imgVal;
					} else val = 0; // outside image boundaries -> no object

					//val = 127-sqrt(wx*wx + wy*wy + wz*wz)*600; // generate sphere as test object

					if (z - prj->CLIP_XY_PLANE_ZSUB < 0) val += (z - prj->CLIP_XY_PLANE_ZSUB)*130; // clip to xy plane

					if (val == 0) break; // speeds up calculation!
				}

				val -= 128;
				slices[z % 2][y*prj->gridSizeX + x] = CLIP(val, -128, 127);

				if (val >= 0) {
					if (wx < lastScanMinX) lastScanMinX = wx;
					if (wx > lastScanMaxX) lastScanMaxX = wx;
					if (wy < lastScanMinY) lastScanMinY = wy;
					if (wy > lastScanMaxY) lastScanMaxY = wy;
					if (wz < lastScanMinZ) lastScanMinZ = wz;
					if (wz > lastScanMaxZ) lastScanMaxZ = wz;
				}

				//if (val == 0) fprintf(logFile, "O"); else fprintf(logFile, (val>0)?"+":"-");
				//fprintf(logFile, "%04i ", val);
			}
			//fprintf(logFile, "\n");
		}

// some 3D filter like occo here? perhaps in future, e.g. to suppress tiny objects

		if (z > 0) {
			fprintf(logFile, "processing slice %i - %i ...", z-1, z);
			prj->mcPtr->processSlice(z-1, slices[(z-1) % 2], slices[z % 2]);
			fprintf(logFile, " finished (num coords: %i, num tris: %i)\n", prj->mcPtr->coords.size(), prj->mcPtr->triangles.size());
		}

		incrementProgressBar();
	}
	prj->mcPtr->terminate(); // to be called after the loop with processSlice
	fprintf(logFile, "number of triangles: %i\n", prj->mcPtr->triangles.size());
	fprintf(logFile, "number of coords: %i\n", prj->mcPtr->coords.size());

	closeProgressBar();

	lastScanMinZ -= prj->resolution * prj->CLIP_XY_PLANE_ZSUB;
	lastScanMaxZ -= prj->resolution * prj->CLIP_XY_PLANE_ZSUB;

	// clean up
	for (int i=0; i<prj->photos.size(); i++) delete img[i];
	delete[] img;

	delete[] slices[0];
	delete[] slices[1];

	fclose(logFile);

	// tell frontend to refresh screen content
	PostMessage(hwnd, WM_USER_REDRAW, GET_IMG_CONTINUE, 0);
}

void ScannerApp::scanFitScanningVolume2LastScan(ScanProject* prj) {
	if (!prj) return;
	if (!prj->mcPtr) return;

	float enlargePercentage = 0.05; // make it a bit bigger than last scan

	float delta = lastScanMaxX - lastScanMinX;
	prj->minX = lastScanMinX - enlargePercentage*delta;
	prj->maxX = lastScanMaxX + enlargePercentage*delta;

	delta = lastScanMaxY - lastScanMinY;
	prj->minY = lastScanMinY - enlargePercentage*delta;
	prj->maxY = lastScanMaxY + enlargePercentage*delta;

	delta = lastScanMaxZ;// - lastScanMinZ;
	//prj->minZ = lastScanMinZ - enlargePercentage*delta;
	prj->maxZ = lastScanMaxZ + enlargePercentage*delta;
}

void ScannerApp::scanSimplifyModel(ScanProject* prj) {
	if (!prj) return;
	if (!prj->mcPtr) return;

/*
	// apply progressive meshes
*/
}

void ScannerApp::scanTexture(ScanProject* prj) {
	if (!prj) return;
	if (!prj->mcPtr) return;

	// open all rgb images and calc cam normals (view direction)
	int photoNum = prj->photos.size();
	MultiResImgPtr* img = new MultiResImgPtr[photoNum];
	VEC4* photoNormals = new VEC4[photoNum];
	for (int i=0; i<photoNum; i++) {
		img[i] = new MultiResImg(prj->photos[i].fileName);

		if (img[i]->getWidth() == 0) {
			MessageBox(hwnd, "Textures couldn't be loaded!", "3DScanner", MB_ICONERROR | MB_OK);
			return;
		}

		// calc cam normals
		VEC4 wcOrigin;
		double d1, d2, d3;
		prj->photos[i].camTrafo.cam2World(0, 0, 0, &d1, &d2, &d3);
		wcOrigin.v[0] = d1;
		wcOrigin.v[1] = d2;
		wcOrigin.v[2] = d3;
		wcOrigin.v[3] = 0;
		resetVec(&photoNormals[i]);
		prj->photos[i].camTrafo.cam2World(0, 0, 1, &d1, &d2, &d3);
		photoNormals[i].v[0] = d1;
		photoNormals[i].v[1] = d2;
		photoNormals[i].v[2] = d3;
		subVec(&photoNormals[i], &wcOrigin);
		normalizeVec3(&photoNormals[i]);
	}

	static const int PROGRESS_INCREMENT = 1000;
	setupProgressBar(hwnd, hInstance, 0, (prj->mcPtr->coords.size()-1) / PROGRESS_INCREMENT);

	// sum up weighted colors for each vertex
	for (int v=0; v<prj->mcPtr->coords.size(); v++) { // all vertices
		VEC4 vn;
		for (int i=0; i<3; i++) vn.v[i] = prj->mcPtr->vertexNormals[v].n[i];
		vn.v[3] = 0;

		float wx = prj->mcPtr->coords[v].c[0];
		float wy = prj->mcPtr->coords[v].c[1];
		float wz = prj->mcPtr->coords[v].c[2];
/*
		wx = (wx / prj->mcPtr->tScaleX) + prj->mcPtr->translX;
		wy = (wy / prj->mcPtr->tScaleY) + prj->mcPtr->translY;
		wz = (wz / prj->mcPtr->tScaleZ) + prj->mcPtr->translZ;

		float gx = wx - 0.5*(prj->gridSizeX-1);
		float gy = wy - 0.5*(prj->gridSizeY-1);
		float gz = wz - prj->CLIP_XY_PLANE_ZSUB;

		wx = prj->resolution*(gx);
		wy = prj->resolution*(gy);
		wz = prj->resolution*(gz);
*/
		float weightSum = 0;
		float rSum = 0;
		float gSum = 0;
		float bSum = 0;
		for (int p=0; p<photoNum; p++) {
			float sm = scalarMultVec3(&vn, &photoNormals[p]);
			if (sm < 0) {

				double imX, imY;
				prj->photos[p].camTrafo.world2Img(1000*wx, 1000*wy, 1000*wz, &imX, &imY);

				// skip if outside img boundaries
				if ((imX >= 0) && (imX < img[p]->getWidth())
					&& (imY >= 0) && (imY < img[p]->getHeight())) {

					// calculation of influence Diameter (for anti aliasing)
					// add distances of image coords for next grid position in each dimension
					// _SLOW_ !!!
					float inflDiam = 0;
					double imTX, imTY;
/*					float iwx = prj->resolution*(gx + 1);
					float iwy = prj->resolution*(gy);
					float iwz = prj->resolution*(gz);
*/
					float iwx = wx + prj->resolution;
					float iwy = wy;
					float iwz = wz;
					prj->photos[p].camTrafo.world2Img(iwx*1000, iwy*1000, iwz*1000, &imTX, &imTY);
					inflDiam += sqrt(sqr(imX - imTX) + sqr(imY - imTY));

/*					iwx = prj->resolution*(gx);
					iwy = prj->resolution*(gy + 1);
					iwz = prj->resolution*(gz);
*/
					iwx = wx;
					iwy = wy + prj->resolution;
					iwz = wz;
					prj->photos[p].camTrafo.world2Img(iwx*1000, iwy*1000, iwz*1000, &imTX, &imTY);
					inflDiam += sqrt(sqr(imX - imTX) + sqr(imY - imTY));

/*					iwx = prj->resolution*(gx);
					iwy = prj->resolution*(gy);
					iwz = prj->resolution*(gz + 1);
*/
					iwx = wx;
					iwy = wy;
					iwz = wz + prj->resolution;
					prj->photos[p].camTrafo.world2Img(iwx*1000, iwy*1000, iwz*1000, &imTX, &imTY);
					inflDiam += sqrt(sqr(imX - imTX) + sqr(imY - imTY));

					inflDiam *= 0.5;

					BYTE tr, tg, tb;
					img[p]->getRGBTriLinIntPolXY(imX, imY, inflDiam, &tr, &tg, &tb);

					float w = sqr(sm); //-sm; // weight

/*					BYTE* cPtr = img[p]->getDataXY((int) imX, (int) imY);
					rSum += *cPtr++ * w;
					gSum += *cPtr++ * w;
					bSum += *cPtr * w;
*/
					rSum += tr * w;
					gSum += tg * w;
					bSum += tb * w;

					weightSum += w;
				}
			}
		}
		if (weightSum == 0) {
			// unknown color --> assign light gray
			prj->mcPtr->vertexColors[v].r = 200;
			prj->mcPtr->vertexColors[v].g = 200;
			prj->mcPtr->vertexColors[v].b = 200;
		} else {
			prj->mcPtr->vertexColors[v].r = (unsigned char) (rSum / weightSum);
			prj->mcPtr->vertexColors[v].g = (unsigned char) (gSum / weightSum);
			prj->mcPtr->vertexColors[v].b = (unsigned char) (bSum / weightSum);
		}

		if (v % PROGRESS_INCREMENT == 0) incrementProgressBar();
	}
	closeProgressBar();

	prj->isTextured = true;

	// clean up
	for (int i=0; i<prj->photos.size(); i++) delete img[i];
	delete[] img;

	// tell frontend to refresh screen content
	PostMessage(hwnd, WM_USER_REDRAW, GET_IMG_CONTINUE, 0);
}

void ScannerApp::configPortTalkInstall() {
	if (MessageBox(hwnd, "This will install PortTalk driver. Do you wish to proceed?", "3DScanner", MB_ICONQUESTION | MB_YESNOCANCEL) != IDYES) return;

	if (InstallPortTalkDriver()) {
		MessageBox(hwnd, "PortTalk driver has successfully been installed.", "3DScanner",  MB_ICONINFORMATION | MB_OK);
		stepperMotorUsed = true;
	} else {
		MessageBox(hwnd, "PortTalk driver couldn't be installed!", "3DScanner",  MB_ICONERROR | MB_OK);
		stepperMotorUsed = false;
	}
}

void ScannerApp::configPortTalkUninstall() {
	if (MessageBox(hwnd, "This will uninstall PortTalk driver. Do you wish to proceed?", "3DScanner", MB_ICONQUESTION | MB_YESNOCANCEL) != IDYES) return;

	if (UninstallPortTalkDriver()) {
		MessageBox(hwnd, "PortTalk driver has successfully been uninstalled.", "3DScanner",  MB_ICONINFORMATION | MB_OK);
		stepperMotorUsed = false;
	} else {
		MessageBox(hwnd, "PortTalk driver couldn't be uninstalled!", "3DScanner",  MB_ICONERROR | MB_OK);
	}
}

void ScannerApp::configSMReset() {
	stepCounter = 0;
}

void ScannerApp::configSMSave1Rot() {
	stepNum360 = stepCounter;
	writeIniFile();
}

void ScannerApp::configSMSetPort(int portAdr) {
	smp.setPortAdr(portAdr);
	writeIniFile();
}

void ScannerApp::configCamSelectSource() {
	camAccess.selectSource(imageAquisitionType);
}

void ScannerApp::configCamOptions() {
MessageBox(hwnd, "is implemented but doesn't work yet - cancelling now...", "3DScanner",  MB_OK);
//	camAccess.configSource(imageAquisitionType);
}

void ScannerApp::configSetIAT(CamAccess::ImageAquisitionType iat) {
	imageAquisitionType = iat;
	writeIniFile();
}

void ScannerApp::configCamShowPreviewWindow() {
/* TODO (#1#): implement with standard aquisition dialog */
MessageBox(hwnd, "not yet implemented", "3DScanner",  MB_OK);
}

void ScannerApp::configSelectCalibrationPattern() {
	MLInputDialog inpDlg(hwnd, hInstance);
	int ind = inpDlg.getIndexFromList((char*) &"Select a calibration pattern.", (char*) &"1 cm\0" "2 cm\0\0");
	if (ind < 0) return;
	switch (ind) {
		case 0: calibrationPattern = CalibrateCam::CAL_PAT_1; break;
		case 1: calibrationPattern = CalibrateCam::CAL_PAT_2; break;
	}
	writeIniFile();
}

// draws OpenGL scene
void ScannerApp::draw() {
	glColor4f(1.0, 1.0, 1.0, 1.0);

	if (project && project->mcPtr) {
		if (!showMesh) glEnable(GL_LIGHTING);
		if (!showMesh) glEnable(GL_CULL_FACE);
glEnable(GL_NORMALIZE);

		if (showSmoothShaded && !showMesh) {
			project->mcPtr->drawSmoothShaded(project->isTextured && showTexture);
		} else {
			if (showMesh) glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
			project->mcPtr->drawFlatShaded(project->isTextured && showTexture);
			if (showMesh) glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
		}

		// draw merge object
		if (appState == APP_STATE_MERGE) {
			glPushMatrix();
			float m[4][4];
			for (int j=0; j<4; j++) for (int i=0; i<4; i++)
				m[i][j] = mergeMatrix.m[j][i];
			glMultMatrixf(&m[0][0]);
			glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
			mergeProject->mcPtr->drawFlatShaded(false);
			glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
			glPopMatrix();
		}

glDisable(GL_NORMALIZE);
		glDisable(GL_LIGHTING);
		glDisable(GL_CULL_FACE);

		if (showNormals) {
			// face normals
			glColor3f(1.0, 0.0, 0.0);
			glBegin(GL_LINES);
			float faceCenter[3];
			for (int t=0; t<project->mcPtr->triangles.size(); t++) {
				for (int v=0; v<3; v++) faceCenter[v] = 0;
				for (int i=0; i<3; i++) {
					int ind = project->mcPtr->triangles[t].t[i];
					for (int v=0; v<3; v++) faceCenter[v] += project->mcPtr->coords[ind].c[v];
				}
				for (int v=0; v<3; v++) faceCenter[v] /= 3;
				glVertex3fv(&faceCenter[0]);
				glVertex3f(faceCenter[0] + project->mcPtr->faceNormals[t].n[0]*project->resolution,
						faceCenter[1] + project->mcPtr->faceNormals[t].n[1]*project->resolution,
						faceCenter[2] + project->mcPtr->faceNormals[t].n[2]*project->resolution);
			}
			glEnd();

			// vertex normals
			glColor3f(0.0, 1.0, 0.0);
			glBegin(GL_LINES);
			for (int v=0; v<project->mcPtr->coords.size(); v++) { // all vertices
				glVertex3fv(&project->mcPtr->coords[v].c[0]);
				glVertex3f(project->mcPtr->coords[v].c[0] + project->mcPtr->vertexNormals[v].n[0]*project->resolution,
						project->mcPtr->coords[v].c[1] + project->mcPtr->vertexNormals[v].n[1]*project->resolution,
						project->mcPtr->coords[v].c[2] + project->mcPtr->vertexNormals[v].n[2]*project->resolution);
			}
			glEnd();
		}
	}

	// grid
	if (showGrid) {
		// brighter lines (meter grid)
		for (int i=-20; i<=20; i++) {
			if (i % 5) glColor3f(0.25, 0.25, 0.4);
				else glColor3f(0.4, 0.4, 0.6);
			glBegin(GL_LINES);
				glVertex3f(i, -21, 0);
				glVertex3f(i, +21, 0);
			glEnd();
			glBegin(GL_LINES);
				glVertex3f(-21, i, 0);
				glVertex3f(+21, i, 0);
			glEnd();
		}
		// vertical line along z axis
		//glBegin(GL_LINES);
		//	glVertex3f(0.0, 0.0, -1.0);
		//	glVertex3f(0.0, 0.0, 1.0);
		//glEnd();

		// blue lines (cm grid)
		for (int i=-30; i<=30; i++) if (i != 0) {
			if (i % 5) glColor3f(0.1, 0.1, 0.4);
				else glColor3f(0.2, 0.2, 0.6);
			glBegin(GL_LINES);
				glVertex3f(i*0.01, -0.31, 0);
				glVertex3f(i*0.01, +0.31, 0);
			glEnd();
			glBegin(GL_LINES);
				glVertex3f(-0.31, i*0.01, 0);
				glVertex3f(+0.31, i*0.01, 0);
			glEnd();
		}
	}

	if (project && project->photos.size() && showViews) {
		glEnable(GL_BLEND);
		glDepthMask(GL_FALSE); // disable writing to depth buffer

		project->visualizeCams();
		if (mergeProject && mergeProject->photos.size()) {
			glPushMatrix();
			float m[4][4];
			for (int j=0; j<4; j++) for (int i=0; i<4; i++)
				m[i][j] = mergeMatrix.m[j][i];
			glMultMatrixf(&m[0][0]);
			mergeProject->visualizeCams();
			glPopMatrix();
		}

		glDisable(GL_BLEND);
		glDepthMask(GL_TRUE); // re-enable writing to depth buffer
	}
}
