/*
 * stepper motor controlling software for own stepper motor interface
 * for 3dscanner project
 *
 * parallel port interface (low level)
 * bipolar stepper motor (two coils)
 *
 * be sure to enable port access before! see PortAccess.*
 *
 *		Marcel Lancelle 2002
 */

#include "StepperMotorParPort.h"
#include "math.h"
#include <windows.h>
#include <winbase.h> // for Sleep

// for parallel port access via '_outp' try one of those
//#include <conio.h> // try this one first (only Win9x)
#include "PortAccess.h" // if your compiler package doesn't have _outp in conio.h

const int ROT_STATE_PORT_1[] = {1, 4, 2, 8, 1, 4, 2, 8}; // normal powered: one winding at a time
const int ROT_STATE_PORT_2[] = {5, 6, 10, 9, 5, 6, 10, 9}; // high powered: two windings at a time
							// (needs twice energy; multiplies torque by about 1.4)
const int ROT_STATE_PORT_3[] = {1, 5, 4, 6, 2, 10, 8, 9}; // half steps

StepperMotorParPort::StepperMotorParPort() {
	powerMode = POWER_OFF;
	port = -1;
	rotationState = 0;

#ifdef USE_PORTTALK
	OpenPortTalk();
#endif // USE_PORTTALK
}

StepperMotorParPort::~StepperMotorParPort() {
#ifdef USE_PORTTALK
	ClosePortTalk();
#endif // USE_PORTTALK
}

// setPower: sets one of the modes above
// think at turning off power after use!
void StepperMotorParPort::setPower(int powerMode) {
	if ((powerMode >= 1) && (powerMode <= 3)) this->powerMode = powerMode;
	else this->powerMode = POWER_OFF;
	updatePort();
}

// setPortAdr: 0x0278, 0x0378 are common adresses for the parallel port
void StepperMotorParPort::setPortAdr(int port) {
	this->port = port;
}

// rotateDirection: rotates in one direction
// when you don't specify a delay think at leaving enough time
// between rotations to stay in valid rotation frequencies
// direction: real direction depends on how motor is connected
// msDelay: delay in ms between two rotation states
// stepNum: number of steps to rotate (negative values allowed)
void StepperMotorParPort::rotate(bool direction) {
	if (direction) rotationState++;
	else rotationState--;
	rotationState &= 7;
	updatePort();
}

void StepperMotorParPort::rotate(bool direction, int msDelay, int stepNum /* = 1 */) {
	if (stepNum < 0) {
		stepNum = -stepNum;
		direction = !direction;
	}

	for (int i=0; i<stepNum; i++) {
		rotate(direction);

		Sleep(msDelay);
	}
}

// rotateAccelerated: linear speed changes
// direction: real direction depends on how motor is connected
// stepNum: number of steps to rotate (negative values allowed)
// minMsDelay: shortest delay specifies max rotation frequency (in ms)
// accMsTime: accelerate during this period (in ms)
//     ______________
//    /              \
//   /                \  time - speed diagram
//  /                  \
//  <->
// accMsTime
void StepperMotorParPort::rotateAccelerated(bool direction, int stepNum, int minMsDelay, int accMsTime) {
	int snum = 0;	// steps already performed
	int td = 0;		// total delay during acceleration period

	int delay = (int) sqrt(accMsTime * minMsDelay);

	if (stepNum < 0) {
		stepNum = -stepNum;
		direction = !direction;
	}

	// speed up
	while (snum < (stepNum-1)/2) {
		rotate(direction);
		snum++;

		if (delay < minMsDelay) delay = minMsDelay; // limit to min delay
		Sleep(delay);
		td += delay;

		// delay for next iteration
		delay = (int) (-td/2 + sqrt(td*td/4 + accMsTime * minMsDelay));
	}

	// slow down
	while (snum < stepNum) {
		rotate(direction);
		snum++;

		// delay
		delay = (int) (-td/2 + sqrt(td*td/4 + accMsTime * minMsDelay));

		if (delay < minMsDelay) delay = minMsDelay; // limit to min delay
		Sleep(delay);
		td -= delay;
	}
}

// updates hardware port corresponding to actual rotation and power states
void StepperMotorParPort::updatePort() {
	if (powerMode == 0) _outp(port, 0);
	else if (powerMode == 1) _outp(port, ROT_STATE_PORT_1[rotationState]);
		else if (powerMode == 2) _outp(port, ROT_STATE_PORT_2[rotationState]);
			else _outp(port, ROT_STATE_PORT_3[rotationState]);
}
