/*
 * Vector (4) and Matrix (4x4) functions
 *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *
 *			00 01 02 03
 * Matrix	10 11 12 13
 *			20 21 22 23
 *			30 31 32 33
 *
 * is declared like that:
 * float a[3][3] = {{00, 01, 02, 03},
 *					 {10, 11, 12, 13},
 *					 {20, 21, 22, 23},
 *					 {30, 31, 32, 33}};
 *
 * access: a[row][column]
 * (d.h.: a[Zeile][Spalte])
 *
 * (a in memory: 00, 01, 02, 03, 10, 11, 12, 13, 20, 21, 22, 23, 30, 31, 32, 33)
 *
 */

#include "math3d.h"
#include <math.h>
#include "mathext.h"
#define deg(x) x*PI/180

void copyVec(VEC4* vSrcPtr, VEC4* vDestPtr) { // vDest = vSrc
	for (int i=0; i<4; i++) vDestPtr->v[i] = vSrcPtr->v[i];
}

void copyMat(MAT4x4* mSrcPtr, MAT4x4* mDestPtr) {
//	for (int i=0; i<16; i++) *((float**) mDestPtr)[i] = *((float**) mSrcPtr)[i];
	for (int i=0; i<4; i++)
	 for (int j=0; j<4; j++)
		mDestPtr->m[i][j] = mSrcPtr->m[i][j];
}

// a = a + b
void addVec(VEC4* a, VEC4* b) {
	for (int i=0; i<4; i++) a->v[i] += b->v[i];
}

// a = a - b
void subVec(VEC4* a, VEC4* b) {
	for (int i=0; i<4; i++) a->v[i] -= b->v[i];
}

// a = a * f
void scaleVec(VEC4* a, float scale) {
	for (int i=0; i<4; i++) a->v[i] *= scale;
}

float getVec3Length(VEC4* a) {
	return sqrt(sqr(a->v[0]) + sqr(a->v[1]) + sqr(a->v[2]));
}

void normalizeVec3(VEC4* a) {
	scaleVec(a, 1/getVec3Length(a));
}

// a * b'
float scalarMultVec3(VEC4* a, VEC4* b) {
	return a->v[0]*b->v[0] + a->v[1]*b->v[1] + a->v[2]*b->v[2];
}

void resetVec(VEC4* vPtr) {
	vPtr->v[0] = 0;
	vPtr->v[1] = 0;
	vPtr->v[2] = 0;
	vPtr->v[3] = 1;
}

void resetMat(MAT4x4* mPtr) {
	for (int i=0; i<4; i++)
	 for (int j=0; j<4; j++)
		mPtr->m[i][j] = (i==j)?1:0;
}

void rotateX(MAT4x4* mPtr, float angle) {  // m = mRot * m
	MAT4x4 org;
	copyMat(mPtr, &org);
	for (int i=0; i<4; i++) {
		mPtr->m[1][i] = cos(deg(angle))*org.m[1][i] - sin(deg(angle))*org.m[2][i];
		mPtr->m[2][i] = sin(deg(angle))*org.m[1][i] + cos(deg(angle))*org.m[2][i];
	}
}

void rotateY(MAT4x4* mPtr, float angle) { // m = mRot * m
	MAT4x4 org;
	copyMat(mPtr, &org);
	for (int i=0; i<4; i++) {
		mPtr->m[0][i] = cos(deg(angle))*org.m[0][i] + sin(deg(angle))*org.m[2][i];
		mPtr->m[2][i] = -sin(deg(angle))*org.m[0][i] + cos(deg(angle))*org.m[2][i];
	}
}

void rotateZ(MAT4x4* mPtr, float angle) { // m = mRot * m
	MAT4x4 org;
	copyMat(mPtr, &org);
	for (int i=0; i<4; i++) {
		mPtr->m[0][i] = cos(deg(angle))*org.m[0][i] - sin(deg(angle))*org.m[1][i];
		mPtr->m[1][i] = sin(deg(angle))*org.m[0][i] + cos(deg(angle))*org.m[1][i];
	}
}

// axis is rotation unit vector
void buildRotationMatrix(MAT4x4* mPtr, VEC4* axis, float angle) {
	//use quaternion to calculate matrix
	float a = (float) cos(deg(angle) / 2);
	float f = (float) sin(deg(angle) / 2);
	float b = axis->v[0] * f;
	float c = axis->v[1] * f;
	float d = axis->v[2] * f;

	mPtr->m[0][0] = 1.0 - 2.0*(c*c + d*d); //a*a + b*b - c*c - d*d;
	mPtr->m[0][1] = 2.0 * (b*c - a*d);
	mPtr->m[0][2] = 2.0 * (b*d + a*c);
	mPtr->m[0][3] = 0.0;

	mPtr->m[1][0] = 2.0f * (b*c + a*d);
	mPtr->m[1][1] = 1.0 - 2.0*(b*b + d*d); //a*a - b*b + c*c - d*d;
	mPtr->m[1][2] = 2.0f * (c*d - a*b);
	mPtr->m[1][3] = 0.0f;

	mPtr->m[2][0] = 2.0f * (b*d - a*c);
	mPtr->m[2][1] = 2.0f * (c*d + a*b);
	mPtr->m[2][2] = 1.0 - 2.0*(b*b + c*c); //a*a - b*b - c*c + d*d;
	mPtr->m[2][3] = 0.0f;

	mPtr->m[3][0] = 0.0;
	mPtr->m[3][1] = 0.0;
	mPtr->m[3][2] = 0.0;
	mPtr->m[3][3] = 1.0;
}

void translate(MAT4x4* mPtr, VEC4* vPtr) {
	for (int i=0; i<3; i++) mPtr->m[i][3] += vPtr->v[i];
}

void multMatVec(MAT4x4* mPtr, VEC4* vPtr) { // vec = mat * vec
	VEC4 org;
	copyVec(vPtr, &org);
	for (int i=0; i<4; i++)
		vPtr->v[i] = mPtr->m[i][0]*org.v[0] + mPtr->m[i][1]*org.v[1] + mPtr->m[i][2]*org.v[2] + mPtr->m[i][3]*org.v[3];
}

void multMatMat(MAT4x4* m1Ptr, MAT4x4* m2Ptr) { // m1 = m1 * m2
	MAT4x4 org;
	copyMat(m1Ptr, &org);
	for (int i=0; i<4; i++)
	 for (int j=0; j<4; j++)
		m1Ptr->m[i][j] = m2Ptr->m[0][j]*org.m[i][0] + m2Ptr->m[1][j]*org.m[i][1] + m2Ptr->m[2][j]*org.m[i][2] + m2Ptr->m[3][j]*org.m[i][3];
}

void multMatMatInv(MAT4x4* m1Ptr, MAT4x4* m2Ptr) { // m1 = m2 * m1
	MAT4x4 org;
	copyMat(m1Ptr, &org);
	for (int i=0; i<4; i++)
	 for (int j=0; j<4; j++)
		m1Ptr->m[i][j] = m2Ptr->m[i][0]*org.m[0][j] + m2Ptr->m[i][1]*org.m[1][j] + m2Ptr->m[i][2]*org.m[2][j] + m2Ptr->m[i][3]*org.m[3][j];
}

void getInvRotMat(MAT4x4* src, MAT4x4* dest) {
	dest->m[0][0] = src->m[0][0];
	dest->m[0][1] = src->m[1][0];
	dest->m[0][2] = src->m[2][0];
	dest->m[0][3] = -src->m[0][0]*src->m[0][3] - src->m[1][0]*src->m[1][3] - src->m[2][0]*src->m[2][3];
	dest->m[1][0] = src->m[0][1];
	dest->m[1][1] = src->m[1][1];
	dest->m[1][2] = src->m[2][1];
	dest->m[1][3] = -src->m[0][1]*src->m[0][3] - src->m[1][1]*src->m[1][3] - src->m[2][1]*src->m[2][3];
	dest->m[2][0] = src->m[0][2];
	dest->m[2][1] = src->m[1][2];
	dest->m[2][2] = src->m[2][2];
	dest->m[2][3] = -src->m[0][2]*src->m[0][3] - src->m[1][2]*src->m[1][3] - src->m[2][2]*src->m[2][3];
	dest->m[3][0] = 0;
	dest->m[3][1] = 0;
	dest->m[3][2] = 0;
	dest->m[3][3] = 1;
}
