/*
 * class for multiresolution images
 *
 * for 3DScanner project
 *
 *		Marcel Lancelle 2002
 */

#include "multiResImg.h"
#include <math.h>

MultiResImg::MultiResImg(char* bmpImgName) {
	MBmpImg* tImg = new MBmpImg(bmpImgName);
	imgNum = 0;
	if (tImg->getWidth() == 0) return;

	imgNum = tImg->getWidth();
	if (tImg->getHeight() > imgNum) imgNum = tImg->getHeight();
	imgNum = (int) ceil(log(imgNum)/log(2));

	img = new MBmpImgPtr[imgNum];

	img[0] = tImg;

	// generate multi resolution images by simple decimation (averaging)
	for (int i=1; i<imgNum; i++) {
		int width = (img[i-1]->getWidth()+1)/2;
		int height = (img[i-1]->getHeight()+1)/2;
		img[i] = new MBmpImg(width, height, tImg->getBitsPerPixel());
		for (int y=0; y<height; y++)
			for (int x=0; x<width; x++) {
				int x1 = x*2; if (x1 >= img[i-1]->getWidth()) x1 = img[i-1]->getWidth() - 1;
				int y1 = y*2; if (y1 >= img[i-1]->getHeight()) y1 = img[i-1]->getHeight() - 1;
				int x2 = x*2 + 1; if (x2 >= img[i-1]->getWidth()) x2 = img[i-1]->getWidth() - 1;
				int y2 = y*2 + 1; if (y2 >= img[i-1]->getHeight()) y2 = img[i-1]->getHeight() - 1;
				switch (tImg->getBitsPerPixel()) {
					case 8: {
						int avr = *img[i-1]->getDataXY(x1, y1);
						avr += *img[i-1]->getDataXY(x2, y1);
						avr += *img[i-1]->getDataXY(x1, y2);
						avr += *img[i-1]->getDataXY(x2, y2);
						*img[i]->getDataXY(x, y) = avr/4;
						break;}
					case 24: {
						BYTE* pc = img[i-1]->getDataXY(x1, y1);
						int avrR = *pc++;
						int avrG = *pc++;
						int avrB = *pc++;
						pc = img[i-1]->getDataXY(x2, y1);
						avrR += *pc++; avrG += *pc++; avrB += *pc++;
						pc = img[i-1]->getDataXY(x1, y2);
						avrR += *pc++; avrG += *pc++; avrB += *pc++;
						pc = img[i-1]->getDataXY(x2, y2);
						avrR += *pc++; avrG += *pc++; avrB += *pc++;

						pc = img[i]->getDataXY(x, y);
						*pc++ = avrR/4;
						*pc++ = avrG/4;
						*pc = avrB/4;
						break;}
				}
			}
	}
}

MultiResImg::~MultiResImg() {
	for (int i=0; i<imgNum; i++) delete img[i];
	delete[] img;
}

int MultiResImg::getWidth() {
	if (imgNum == 0) return 0;
	return img[0]->getWidth();
}

int MultiResImg::getHeight() {
	if (imgNum == 0) return 0;
	return img[0]->getHeight();
}

int MultiResImg::getBitsPerPixel() {
	if (imgNum == 0) return 0;
	return img[0]->getBitsPerPixel();
}

// trilinear interpolation

int MultiResImg::getGrayTriLinIntPolXY(float x, float y, float influenceDiam) {
	if (influenceDiam <= 1) return img[0]->getGrayLinIntPolXY(x, y);

	float ldDiam = log(influenceDiam) / log(2);

	int flDiam = (int) floor(ldDiam); // img number of the bigger image (smaller image is flDiam + 1)
	if (ldDiam >= imgNum-1) return *img[imgNum-1]->getDataXY(0, 0);

	float weight = ldDiam - flDiam;

	float f = 1.0 / pow(2, flDiam); // coordinate scaling for bigger image
	return (int) (img[flDiam]->getGrayLinIntPolXY(x * f, y * f) * (1-weight)
				+ img[flDiam + 1]->getGrayLinIntPolXY(x * 0.5 * f, y * 0.5 * f) * weight);
}

void MultiResImg::getRGBTriLinIntPolXY(float x, float y, float influenceDiam, BYTE* r, BYTE* g, BYTE* b) {
	if (influenceDiam <= 1) {
		img[0]->getRGBLinIntPolXY(x, y, r, g, b);
		return;
	}

	float ldDiam = log(influenceDiam) / log(2);

	int flDiam = (int) floor(ldDiam); // img number of the bigger image (smaller image is flDiam + 1)
	if (ldDiam >= imgNum-1) {
		BYTE* cPtr = img[imgNum-1]->getDataXY(0, 0);
		*r = *cPtr++;
		*g = *cPtr++;
		*b = *cPtr;
		return;
	}

	float weight = ldDiam - flDiam;

	float f = 1.0 / pow(2, flDiam); // coordinate scaling for bigger image

	BYTE r1, g1, b1, r2, g2, b2;
	img[flDiam]->getRGBLinIntPolXY(x * f, y * f, &r1, &g1, &b1);
	img[flDiam + 1]->getRGBLinIntPolXY(x * 0.5 * f, y * 0.5 * f, &r2, &g2, &b2);
	*r = (int) (r1 * (1-weight) + r2 * weight);
	*g = (int) (g1 * (1-weight) + g2 * weight);
	*b = (int) (b1 * (1-weight) + b2 * weight);
}
