/*
 * segmentation class for separating the object from the monochrome background
 * for 3dscanner project
 *
 *		Marcel Lancelle 2002
 */

#include "segmentation.h"
#include <math.h>
#include "mathext.h"
//#include <stdio.h> // for debugging only

#define SHADOW_LUMINOSITY_DIFF 25 // estimated shadow regions are darker by this value
#define SCALE_DIFF_IMG 400 // intensity of difference image before thresholding
/*static*/ int Segmentation::THRESHOLD_BG = 20; // all pixels below belong to background (and can't become object pixel any more)
/*static*/ int Segmentation::THRESHOLD_OBJ = 60; // all pixels below belong to object (and can't become background pixel any more)
#define THRESHOLD_BETWEEN 70 // expected threshold (important for preparing watershed image)
#define DISTANCE_IMG_BORDER 4 // image border will not be marked as object within this distance (cheap cameras somtimes have bad borders)

/* static */ bool Segmentation::debugOutput = false;

void getHSLModifiedCoords(int h, int s, int l, float* x, float* y, float* z) {
	float norm = 1 / 255.0;
	float dist = sqrt((128-abs(l-127)) * 2 * norm) * s * norm;
	*x = dist * cos(h*norm*2*PI);
	*y = dist * sin(h*norm*2*PI);
	*z = 0.5*(2*norm*l - 1);
}

/*static*/ void Segmentation::setDebugOutput(bool debugOutput) {
	Segmentation::debugOutput = debugOutput;
}

/*static*/ void Segmentation::segment(MBmpImg* photo, MBmpImg* mask) {
	int width = photo->getWidth();
	int height = photo->getHeight();
	MBmpImg img2(width, height, 24);

	photo->RGB2HSL(&img2);
	img2.medianRGB3x3(photo);

	MBmpImg gradImg(width, height, 8);
	photo->sobelRGB2Gray(&gradImg, 1.0);

	// estimate background color: take color near each corner
	static const int BG_COL_NUM = 8;
	float bgX[BG_COL_NUM], bgY[BG_COL_NUM], bgZ[BG_COL_NUM];
	for (int i=0; i<BG_COL_NUM/2; i++) {
		int borderDist = ((width > 50) && (height > 50)) ? 10 : 0;
		BYTE* bgDatPtr = photo->getDataXY((i/2)?borderDist:width-borderDist, (i%2)?borderDist:height-borderDist);
		int bgH = *bgDatPtr++;
		int bgS = *bgDatPtr++;
		int bgL = *bgDatPtr;
		getHSLModifiedCoords(bgH, bgS, bgL, &bgX[i], &bgY[i], &bgZ[i]);
		// estimate same color more darker for shadow regions
		bgL = CLIP(bgL - SHADOW_LUMINOSITY_DIFF, 0, 255);
		getHSLModifiedCoords(bgH, bgS, bgL, &bgX[i+BG_COL_NUM/2], &bgY[i+BG_COL_NUM/2], &bgZ[i+BG_COL_NUM/2]);
	}

	MBmpImg beforeTresholdImg(width, height, 8);

	/* TODO (#1#): for distance perhaps discrete 3d
	               distance map? */

	// generate color difference image to background color (HSL model)
	for (int y=0; y<height; y++)
		for (int x=0; x<width; x++) { // all pixel
			BYTE* datPtr = photo->getDataXY(x, y);

			BYTE h = *datPtr++;
			BYTE s = *datPtr++;
			BYTE l = *datPtr;

			// calc minimal distance (background color to pixel color) in HSL color space
			float cx, cy, cz;
			getHSLModifiedCoords(h, s, l, &cx, &cy, &cz);

			int val = 255;
			for (int i=0; i<BG_COL_NUM; i++) {
				int v = (int) (SCALE_DIFF_IMG * sqrt(sqr(bgX[i] - cx) + sqr(bgY[i] - cy) + sqr(bgZ[i] - cz)));
				if (v < val) val = v;
			}
/*
datPtr = img2.getDataXY(x, y);
*datPtr++ = (int) CLIP(128 + 200*(cx - bgX[0]), 0, 255);
*datPtr++ = (int) CLIP(128 + 200*(cy - bgY[0]), 0, 255);
*datPtr = (int) CLIP(128 + 200*(cz - bgZ[0]), 0, 255);
*/
*beforeTresholdImg.getDataXY(x, y) = CLIP(val, 0, 255);

			// threshold for background
			BYTE tVal = 128;
			if (val < THRESHOLD_BG) tVal = 0;
			if (val > THRESHOLD_OBJ) {
				tVal = 255;
				if ((x < DISTANCE_IMG_BORDER) || (x >= width - DISTANCE_IMG_BORDER)
					|| (y < DISTANCE_IMG_BORDER) || (y >= height - DISTANCE_IMG_BORDER)) tVal = 128;
			}
			*mask->getDataXY(x, y) = tVal;
		}

	if (debugOutput) beforeTresholdImg.write2BmpFile("segtest before threshold.bmp");
//img2.write2BmpFile("segtest dx dy dz.bmp");

	gradImg.write2BmpFile("segtest gradImg.bmp");
	mask->write2BmpFile("segtest label.bmp");

	// gradImg contains the gradient
	// mask contains 3 regions: 0: background, 128: undef, 255: object

	// prepare gradient image for watersheds
	// - use background and object regions of mask as markers
	// - not marker pixels equal 0 are incremented for speed reasons (see below)

	for (int y=0; y<height; y++)
		for (int x=0; x<width; x++) { // all pixel
			if (*mask->getDataXY(x, y) != 128) *gradImg.getDataXY(x, y) = 0;
			else {
				int val = *beforeTresholdImg.getDataXY(x, y);
				if (val <= THRESHOLD_BETWEEN) val = (int) (sqr((THRESHOLD_BETWEEN-val)/((float) THRESHOLD_BETWEEN-THRESHOLD_BG)) * 50.0);
					else val = (int) (sqr((val-THRESHOLD_BETWEEN)/((float) THRESHOLD_OBJ-THRESHOLD_BETWEEN)) * 100.0);
//				*gradImg.getDataXY(x, y) = CLIP(*gradImg.getDataXY(x, y) - 255*sqrt(val/255.0), 1, 255);
//				*gradImg.getDataXY(x, y) = CLIP(*gradImg.getDataXY(x, y) - val, 1, 255);
				*gradImg.getDataXY(x, y) = CLIP(*gradImg.getDataXY(x, y) - val, 0, 254) + 1; // +1: make local minima > 0
//				*gradImg.getDataXY(x, y) = CLIP(*gradImg.getDataXY(x, y) * 0.5*(255-val), 1, 255);
			}
		}

	if (debugOutput) gradImg.write2BmpFile("segtest gradImg marked.bmp");


	// do watershed transform with flooding simulation; use only initial minima, no new ones at higher levels
	// --> minima imposition technique removes these local minima (in this case: does hardly change anything)
	//           - local minimal allready have a gray level > 0
    //           - now do morphologic reconstruction / algebraic closing (by repeated geodaetic erosion of label image (minImpGradImg) by modified gradient image (gradImg))

	MBmpImg minImpGradImg(width, height, 8);
	for (int y=0; y<height; y++)
		for (int x=0; x<width; x++) { // all pixel
			*minImpGradImg.getDataXY(x, y) = (*mask->getDataXY(x, y) == 128) ? 255 : 0;
		}

	// iterative geodaetic erosion (with 3x3 square as SE); geaodaetic erosion: max(image, eroded mask image) for each pixel
	int modifiedPixelNum;
	do {
		modifiedPixelNum = 0;
		// forward scanning
		for (int y=0; y<height; y++)
			for (int x=0; x<width; x++) { // all pixel
				int min = *minImpGradImg.getDataXY(x, y);
				// test backward neighbors
				int v;
				v = *minImpGradImg.getDataXY(CLIP(x-1, 0, width-1), CLIP(y-1, 0, height-1));
				if (v < min) min = v;
				v = *minImpGradImg.getDataXY(x, CLIP(y-1, 0, height-1));
				if (v < min) min = v;
				v = *minImpGradImg.getDataXY(CLIP(x+1, 0, width-1), CLIP(y-1, 0, height-1));
				if (v < min) min = v;
				v = *minImpGradImg.getDataXY(CLIP(x-1, 0, width-1), y);
				if (v < min) min = v;

				int s = *gradImg.getDataXY(x, y);
				int max = (s > min) ? s : min;
				if (*minImpGradImg.getDataXY(x, y) != max) {
					modifiedPixelNum++;
					*minImpGradImg.getDataXY(x, y) = max; // same image can be used (is even faster)
				}
			}

		// backward scanning
		for (int y=height-1; y>=0; y--)
			for (int x=width-1; x>=0; x--) { // all pixel
				int min = *minImpGradImg.getDataXY(x, y);
				// test forward neighbors
				int v;
				v = *minImpGradImg.getDataXY(CLIP(x+1, 0, width-1), y);
				if (v < min) min = v;
				v = *minImpGradImg.getDataXY(CLIP(x-1, 0, width-1), CLIP(y+1, 0, height-1));
				if (v < min) min = v;
				v = *minImpGradImg.getDataXY(x, CLIP(y+1, 0, height-1));
				if (v < min) min = v;
				v = *minImpGradImg.getDataXY(CLIP(x+1, 0, width-1), CLIP(y+1, 0, height-1));
				if (v < min) min = v;

				int s = *gradImg.getDataXY(x, y);
				int max = (s > min) ? s : min;
				if (*minImpGradImg.getDataXY(x, y) != max) {
					modifiedPixelNum++;
					*minImpGradImg.getDataXY(x, y) = max; // same image can be used (is even faster)
				}
			}
	} while (modifiedPixelNum);

	if (debugOutput) minImpGradImg.write2BmpFile("segtest gradImg marked min imp.bmp");

	// for speeding up pixel access sort by gray value

	//calculate cumulative histogram
	int histo[256];
	for (int i=0; i<256; i++) histo[i] = 0;
	for (int y=0; y<height; y++)
		for (int x=0; x<width; x++) histo[*minImpGradImg.getDataXY(x, y)]++;
	int cumulHisto[256];
	for (int i=0; i<256; i++) cumulHisto[i] = histo[i];
	for (int i=1; i<256; i++) cumulHisto[i] += cumulHisto[i-1];
	// additional variables for sorting
	int pixLevAdr[256];
	pixLevAdr[0] = 0;
	for (int i=1; i<256; i++) pixLevAdr[i] = cumulHisto[i-1];
	int* sortedGrayValuePtrs = new int[width * height];
	// sort
	for (int y=0; y<height; y++)
		for (int x=0; x<width; x++) {
		BYTE* ptr = minImpGradImg.getDataXY(x, y);
		sortedGrayValuePtrs[pixLevAdr[*ptr]] = x+y*width;
		pixLevAdr[*ptr]++;
	}
	// reset pixLevAdr
	pixLevAdr[0] = 0;
	for (int i=1; i<256; i++) pixLevAdr[i] = cumulHisto[i-1];

	MBmpImg tmpMask(width, height, 8);

	for (int y=0; y<height; y++)
		for (int x=0; x<width; x++) *tmpMask.getDataXY(x, y) = *mask->getDataXY(x, y);
	MBmpImg* dSource = mask;
	MBmpImg* dDest = &tmpMask;

	for (int level=0; level<256; level++) {
//printf("l: %i %i\n", histo[level], cumulHisto[level]);
		int modifiedPixNum;
		do {
//printf("_");
			modifiedPixNum = 0;
			// don't take level 0 pixel --> speeds up a lot but is not really correct
//			for (int li=0; li<histo[level]; li++) { // all pixel with gray value = level
			for (int li=0; li<cumulHisto[level]-histo[0]; li++) { // all pixel with gray value = level
//				int adrInd = sortedGrayValuePtrs[pixLevAdr[level] + li];
				int adrInd = sortedGrayValuePtrs[histo[0] + li];
				if (*(BYTE*) (dDest->getData() + adrInd) == 128) { // all pixel without label
					int x = adrInd % width;
					int y = adrInd / width;
					int numBg = 0;
					int numObj = 0;

					for (int j=-1; j<2; j++) for (int i=-1; i<2; i++) if ((i != 0) || (j != 0)) {
						BYTE l = *dSource->getDataXY(CLIP(x+i, 0, width-1), CLIP(y+j, 0, height-1));
						if (l == 0) numBg++;
						if (l == 255) numObj++;
					}

					if ((numBg > 0) && (numObj == 0)) {
// adrInd benutzen
						*dDest->getDataXY(x, y) = 0;
						modifiedPixNum++;
					}
					if ((numBg == 0) && (numObj > 0)) {
						*dDest->getDataXY(x, y) = 255;
						modifiedPixNum++;
					}
				}
			}
			MBmpImg* tmpPtr = dSource;
			dSource = dDest;
			dDest = tmpPtr;
//printf(" %i ", modifiedPixNum);

		} while (modifiedPixNum);
	}

	delete[] sortedGrayValuePtrs;
/*
	// last level (255) : treat all pixel <= 255 to fill local minima
	int modifiedPixNum;
	do {
printf(".");
		modifiedPixNum = 0;
		for (int y=0; y<height; y++)
			for (int x=0; x<width; x++) if (*dDest->getDataXY(x, y) == 128) { // all pixel with value <= level and without label
				int numBg = 0;
				int numObj = 0;
				for (int j=-1; j<2; j++) for (int i=-1; i<2; i++) if ((i != 0) || (j != 0)) {
					BYTE l = *dSource->getDataXY(CLIP(x+i, 0, width-1), CLIP(y+j, 0, height-1));
					if (l == 0) numBg++;
					if (l == 255) numObj++;
				}

				if ((numBg > 0) && (numObj == 0)) {
					*dDest->getDataXY(x, y) = 0;
					modifiedPixNum++;
				}
				if ((numBg == 0) && (numObj > 0)) {
					*dDest->getDataXY(x, y) = 255;
					modifiedPixNum++;
				}
			}
		MBmpImg* tmpPtr = dSource;
		dSource = dDest;
		dDest = tmpPtr;
	} while (modifiedPixNum);
*/
	mask->blurGaussian121(&gradImg);
}
